"""File b05latent_choice_full.py

Choice model with the latent variable.
Mixture of logit.
Measurement equation for the indicators.
Maximum likelihood (full information) estimation.

:author: Michel Bierlaire, EPFL
:date: Thu Apr 13 18:02:26 2023
"""

import sys
import biogeme.logging as blog
import biogeme.biogeme as bio
import biogeme.exceptions as excep
from biogeme import models
import biogeme.distributions as dist
import biogeme.results as res
from biogeme.expressions import (
    Beta,
    log,
    RandomVariable,
    Integrate,
    Elem,
    bioNormalCdf,
    exp,
)

from optima import (
    database,
    age_65_more,
    formulaIncome,
    moreThanOneCar,
    moreThanOneBike,
    individualHouse,
    male,
    haveChildren,
    haveGA,
    highEducation,
    TimePT,
    MarginalCostPT,
    TimeCar,
    CostCarCHF,
    distance_km,
    TripPurpose,
    WaitingTimePT,
    Envir01,
    Envir02,
    Envir03,
    Mobil11,
    Mobil14,
    Mobil16,
    Mobil17,
    Choice,
)

logger = blog.get_screen_logger(level=blog.INFO)
logger.info('Example b05latent_choice_full.py')


# Read the estimates from the structural equation estimation
FILENAME = 'b02one_latent_ordered'
try:
    structResults = res.bioResults(pickleFile=f'{FILENAME}.pickle')
except excep.BiogemeError:
    print(
        f'Run first the script {FILENAME}.py in order to generate the '
        f'file {FILENAME}.pickle.'
    )
    sys.exit()
structBetas = structResults.getBetaValues()

# Coefficients

coef_intercept = Beta('coef_intercept', structBetas['coef_intercept'], None, None, 0)
coef_age_65_more = Beta(
    'coef_age_65_more', structBetas['coef_age_65_more'], None, None, 0
)
coef_haveGA = Beta('coef_haveGA', structBetas['coef_haveGA'], None, None, 0)

coef_moreThanOneCar = Beta(
    'coef_moreThanOneCar', structBetas['coef_moreThanOneCar'], None, None, 0
)
coef_moreThanOneBike = Beta(
    'coef_moreThanOneBike', structBetas['coef_moreThanOneBike'], None, None, 0
)
coef_individualHouse = Beta(
    'coef_individualHouse', structBetas['coef_individualHouse'], None, None, 0
)
coef_male = Beta('coef_male', structBetas['coef_male'], None, None, 0)
coef_haveChildren = Beta(
    'coef_haveChildren', structBetas['coef_haveChildren'], None, None, 0
)
coef_highEducation = Beta(
    'coef_highEducation', structBetas['coef_highEducation'], None, None, 0
)

# Latent variable: structural equation

# Define a random parameter, normally distributed, designed to be used
# for numerical integration
omega = RandomVariable('omega')
density = dist.normalpdf(omega)
sigma_s = Beta('sigma_s', 1, None, None, 0)

CARLOVERS = (
    coef_intercept
    + coef_age_65_more * age_65_more
    + formulaIncome
    + coef_moreThanOneCar * moreThanOneCar
    + coef_moreThanOneBike * moreThanOneBike
    + coef_individualHouse * individualHouse
    + coef_male * male
    + coef_haveChildren * haveChildren
    + coef_haveGA * haveGA
    + coef_highEducation * highEducation
    + sigma_s * omega
)

# Measurement equations

INTER_Envir01 = Beta('INTER_Envir01', 0, None, None, 1)
INTER_Envir02 = Beta('INTER_Envir02', 0, None, None, 0)
INTER_Envir03 = Beta('INTER_Envir03', 0, None, None, 0)
INTER_Mobil11 = Beta('INTER_Mobil11', 0, None, None, 0)
INTER_Mobil14 = Beta('INTER_Mobil14', 0, None, None, 0)
INTER_Mobil16 = Beta('INTER_Mobil16', 0, None, None, 0)
INTER_Mobil17 = Beta('INTER_Mobil17', 0, None, None, 0)

B_Envir01_F1 = Beta('B_Envir01_F1', -1, None, None, 1)
B_Envir02_F1 = Beta('B_Envir02_F1', -1, None, None, 0)
B_Envir03_F1 = Beta('B_Envir03_F1', 1, None, None, 0)
B_Mobil11_F1 = Beta('B_Mobil11_F1', 1, None, None, 0)
B_Mobil14_F1 = Beta('B_Mobil14_F1', 1, None, None, 0)
B_Mobil16_F1 = Beta('B_Mobil16_F1', 1, None, None, 0)
B_Mobil17_F1 = Beta('B_Mobil17_F1', 1, None, None, 0)

MODEL_Envir01 = INTER_Envir01 + B_Envir01_F1 * CARLOVERS
MODEL_Envir02 = INTER_Envir02 + B_Envir02_F1 * CARLOVERS
MODEL_Envir03 = INTER_Envir03 + B_Envir03_F1 * CARLOVERS
MODEL_Mobil11 = INTER_Mobil11 + B_Mobil11_F1 * CARLOVERS
MODEL_Mobil14 = INTER_Mobil14 + B_Mobil14_F1 * CARLOVERS
MODEL_Mobil16 = INTER_Mobil16 + B_Mobil16_F1 * CARLOVERS
MODEL_Mobil17 = INTER_Mobil17 + B_Mobil17_F1 * CARLOVERS

SIGMA_STAR_Envir01 = Beta('SIGMA_STAR_Envir01', 1, 1.0e-5, None, 1)
SIGMA_STAR_Envir02 = Beta('SIGMA_STAR_Envir02', 1, 1.0e-5, None, 0)
SIGMA_STAR_Envir03 = Beta('SIGMA_STAR_Envir03', 1, 1.0e-5, None, 0)
SIGMA_STAR_Mobil11 = Beta('SIGMA_STAR_Mobil11', 1, 1.0e-5, None, 0)
SIGMA_STAR_Mobil14 = Beta('SIGMA_STAR_Mobil14', 1, 1.0e-5, None, 0)
SIGMA_STAR_Mobil16 = Beta('SIGMA_STAR_Mobil16', 1, 1.0e-5, None, 0)
SIGMA_STAR_Mobil17 = Beta('SIGMA_STAR_Mobil17', 1, 1.0e-5, None, 0)


delta_1 = Beta('delta_1', 0.1, 1.0e-5, None, 0)
delta_2 = Beta('delta_2', 0.2, 1.0e-5, None, 0)
tau_1 = -delta_1 - delta_2
tau_2 = -delta_1
tau_3 = delta_1
tau_4 = delta_1 + delta_2

Envir01_tau_1 = (tau_1 - MODEL_Envir01) / SIGMA_STAR_Envir01
Envir01_tau_2 = (tau_2 - MODEL_Envir01) / SIGMA_STAR_Envir01
Envir01_tau_3 = (tau_3 - MODEL_Envir01) / SIGMA_STAR_Envir01
Envir01_tau_4 = (tau_4 - MODEL_Envir01) / SIGMA_STAR_Envir01
IndEnvir01 = {
    1: bioNormalCdf(Envir01_tau_1),
    2: bioNormalCdf(Envir01_tau_2) - bioNormalCdf(Envir01_tau_1),
    3: bioNormalCdf(Envir01_tau_3) - bioNormalCdf(Envir01_tau_2),
    4: bioNormalCdf(Envir01_tau_4) - bioNormalCdf(Envir01_tau_3),
    5: 1 - bioNormalCdf(Envir01_tau_4),
    6: 1.0,
    -1: 1.0,
    -2: 1.0,
}

P_Envir01 = Elem(IndEnvir01, Envir01)

Envir02_tau_1 = (tau_1 - MODEL_Envir02) / SIGMA_STAR_Envir02
Envir02_tau_2 = (tau_2 - MODEL_Envir02) / SIGMA_STAR_Envir02
Envir02_tau_3 = (tau_3 - MODEL_Envir02) / SIGMA_STAR_Envir02
Envir02_tau_4 = (tau_4 - MODEL_Envir02) / SIGMA_STAR_Envir02
IndEnvir02 = {
    1: bioNormalCdf(Envir02_tau_1),
    2: bioNormalCdf(Envir02_tau_2) - bioNormalCdf(Envir02_tau_1),
    3: bioNormalCdf(Envir02_tau_3) - bioNormalCdf(Envir02_tau_2),
    4: bioNormalCdf(Envir02_tau_4) - bioNormalCdf(Envir02_tau_3),
    5: 1 - bioNormalCdf(Envir02_tau_4),
    6: 1.0,
    -1: 1.0,
    -2: 1.0,
}

P_Envir02 = Elem(IndEnvir02, Envir02)

Envir03_tau_1 = (tau_1 - MODEL_Envir03) / SIGMA_STAR_Envir03
Envir03_tau_2 = (tau_2 - MODEL_Envir03) / SIGMA_STAR_Envir03
Envir03_tau_3 = (tau_3 - MODEL_Envir03) / SIGMA_STAR_Envir03
Envir03_tau_4 = (tau_4 - MODEL_Envir03) / SIGMA_STAR_Envir03
IndEnvir03 = {
    1: bioNormalCdf(Envir03_tau_1),
    2: bioNormalCdf(Envir03_tau_2) - bioNormalCdf(Envir03_tau_1),
    3: bioNormalCdf(Envir03_tau_3) - bioNormalCdf(Envir03_tau_2),
    4: bioNormalCdf(Envir03_tau_4) - bioNormalCdf(Envir03_tau_3),
    5: 1 - bioNormalCdf(Envir03_tau_4),
    6: 1.0,
    -1: 1.0,
    -2: 1.0,
}

P_Envir03 = Elem(IndEnvir03, Envir03)

Mobil11_tau_1 = (tau_1 - MODEL_Mobil11) / SIGMA_STAR_Mobil11
Mobil11_tau_2 = (tau_2 - MODEL_Mobil11) / SIGMA_STAR_Mobil11
Mobil11_tau_3 = (tau_3 - MODEL_Mobil11) / SIGMA_STAR_Mobil11
Mobil11_tau_4 = (tau_4 - MODEL_Mobil11) / SIGMA_STAR_Mobil11
IndMobil11 = {
    1: bioNormalCdf(Mobil11_tau_1),
    2: bioNormalCdf(Mobil11_tau_2) - bioNormalCdf(Mobil11_tau_1),
    3: bioNormalCdf(Mobil11_tau_3) - bioNormalCdf(Mobil11_tau_2),
    4: bioNormalCdf(Mobil11_tau_4) - bioNormalCdf(Mobil11_tau_3),
    5: 1 - bioNormalCdf(Mobil11_tau_4),
    6: 1.0,
    -1: 1.0,
    -2: 1.0,
}

P_Mobil11 = Elem(IndMobil11, Mobil11)

Mobil14_tau_1 = (tau_1 - MODEL_Mobil14) / SIGMA_STAR_Mobil14
Mobil14_tau_2 = (tau_2 - MODEL_Mobil14) / SIGMA_STAR_Mobil14
Mobil14_tau_3 = (tau_3 - MODEL_Mobil14) / SIGMA_STAR_Mobil14
Mobil14_tau_4 = (tau_4 - MODEL_Mobil14) / SIGMA_STAR_Mobil14
IndMobil14 = {
    1: bioNormalCdf(Mobil14_tau_1),
    2: bioNormalCdf(Mobil14_tau_2) - bioNormalCdf(Mobil14_tau_1),
    3: bioNormalCdf(Mobil14_tau_3) - bioNormalCdf(Mobil14_tau_2),
    4: bioNormalCdf(Mobil14_tau_4) - bioNormalCdf(Mobil14_tau_3),
    5: 1 - bioNormalCdf(Mobil14_tau_4),
    6: 1.0,
    -1: 1.0,
    -2: 1.0,
}

P_Mobil14 = Elem(IndMobil14, Mobil14)

Mobil16_tau_1 = (tau_1 - MODEL_Mobil16) / SIGMA_STAR_Mobil16
Mobil16_tau_2 = (tau_2 - MODEL_Mobil16) / SIGMA_STAR_Mobil16
Mobil16_tau_3 = (tau_3 - MODEL_Mobil16) / SIGMA_STAR_Mobil16
Mobil16_tau_4 = (tau_4 - MODEL_Mobil16) / SIGMA_STAR_Mobil16
IndMobil16 = {
    1: bioNormalCdf(Mobil16_tau_1),
    2: bioNormalCdf(Mobil16_tau_2) - bioNormalCdf(Mobil16_tau_1),
    3: bioNormalCdf(Mobil16_tau_3) - bioNormalCdf(Mobil16_tau_2),
    4: bioNormalCdf(Mobil16_tau_4) - bioNormalCdf(Mobil16_tau_3),
    5: 1 - bioNormalCdf(Mobil16_tau_4),
    6: 1.0,
    -1: 1.0,
    -2: 1.0,
}

P_Mobil16 = Elem(IndMobil16, Mobil16)

Mobil17_tau_1 = (tau_1 - MODEL_Mobil17) / SIGMA_STAR_Mobil17
Mobil17_tau_2 = (tau_2 - MODEL_Mobil17) / SIGMA_STAR_Mobil17
Mobil17_tau_3 = (tau_3 - MODEL_Mobil17) / SIGMA_STAR_Mobil17
Mobil17_tau_4 = (tau_4 - MODEL_Mobil17) / SIGMA_STAR_Mobil17
IndMobil17 = {
    1: bioNormalCdf(Mobil17_tau_1),
    2: bioNormalCdf(Mobil17_tau_2) - bioNormalCdf(Mobil17_tau_1),
    3: bioNormalCdf(Mobil17_tau_3) - bioNormalCdf(Mobil17_tau_2),
    4: bioNormalCdf(Mobil17_tau_4) - bioNormalCdf(Mobil17_tau_3),
    5: 1 - bioNormalCdf(Mobil17_tau_4),
    6: 1.0,
    -1: 1.0,
    -2: 1.0,
}

P_Mobil17 = Elem(IndMobil17, Mobil17)

# Choice model
# Read the estimates from the sequential estimation, and use
# them as starting values

FILENAME = 'b04latent_choice_seq'
try:
    choiceResults = res.bioResults(pickleFile=f'{FILENAME}.pickle')
except excep.BiogemeError:
    print(
        f'Run first the script {FILENAME}.py in order to generate the '
        f'file {FILENAME}.pickle.'
    )
    sys.exit()

choiceBetas = choiceResults.getBetaValues()

ASC_CAR = Beta('ASC_CAR', choiceBetas['ASC_CAR'], None, None, 0)
ASC_PT = Beta('ASC_PT', 0, None, None, 1)
ASC_SM = Beta('ASC_SM', choiceBetas['ASC_SM'], None, None, 0)
BETA_COST_HWH = Beta('BETA_COST_HWH', choiceBetas['BETA_COST_HWH'], None, None, 0)
BETA_COST_OTHER = Beta('BETA_COST_OTHER', choiceBetas['BETA_COST_OTHER'], None, None, 0)
BETA_DIST = Beta('BETA_DIST', choiceBetas['BETA_DIST'], None, None, 0)
BETA_TIME_CAR_REF = Beta(
    'BETA_TIME_CAR_REF', choiceBetas['BETA_TIME_CAR_REF'], None, 0, 0
)
BETA_TIME_CAR_CL = Beta(
    'BETA_TIME_CAR_CL', choiceBetas['BETA_TIME_CAR_CL'], None, None, 0
)
BETA_TIME_PT_REF = Beta('BETA_TIME_PT_REF', choiceBetas['BETA_TIME_PT_REF'], None, 0, 0)
BETA_TIME_PT_CL = Beta('BETA_TIME_PT_CL', choiceBetas['BETA_TIME_PT_CL'], None, None, 0)
BETA_WAITING_TIME = Beta(
    'BETA_WAITING_TIME', choiceBetas['BETA_WAITING_TIME'], None, None, 0
)

TimePT_scaled = database.DefineVariable('TimePT_scaled', TimePT / 200)
TimeCar_scaled = database.DefineVariable('TimeCar_scaled', TimeCar / 200)
MarginalCostPT_scaled = database.DefineVariable(
    'MarginalCostPT_scaled', MarginalCostPT / 10
)
CostCarCHF_scaled = database.DefineVariable('CostCarCHF_scaled', CostCarCHF / 10)
distance_km_scaled = database.DefineVariable('distance_km_scaled', distance_km / 5)
PurpHWH = database.DefineVariable('PurpHWH', TripPurpose == 1)
PurpOther = database.DefineVariable('PurpOther', TripPurpose != 1)

# Definition of utility functions:

BETA_TIME_PT = BETA_TIME_PT_REF * exp(BETA_TIME_PT_CL * CARLOVERS)

V0 = (
    ASC_PT
    + BETA_TIME_PT * TimePT_scaled
    + BETA_WAITING_TIME * WaitingTimePT
    + BETA_COST_HWH * MarginalCostPT_scaled * PurpHWH
    + BETA_COST_OTHER * MarginalCostPT_scaled * PurpOther
)

BETA_TIME_CAR = BETA_TIME_CAR_REF * exp(BETA_TIME_CAR_CL * CARLOVERS)

V1 = (
    ASC_CAR
    + BETA_TIME_CAR * TimeCar_scaled
    + BETA_COST_HWH * CostCarCHF_scaled * PurpHWH
    + BETA_COST_OTHER * CostCarCHF_scaled * PurpOther
)

V2 = ASC_SM + BETA_DIST * distance_km_scaled

# Associate utility functions with the numbering of alternatives
V = {0: V0, 1: V1, 2: V2}

# Conditional to omega, we have a logit model (called the kernel) for
# the choice
condprob = models.logit(V, None, Choice)

# Conditional to omega, we have the product of ordered probit for the
# indicators.
condlike = (
    P_Envir01
    * P_Envir02
    * P_Envir03
    * P_Mobil11
    * P_Mobil14
    * P_Mobil16
    * P_Mobil17
    * condprob
)

# We integrate over omega using numerical integration
loglike = log(Integrate(condlike * density, 'omega'))

# Create the Biogeme object
the_biogeme = bio.BIOGEME(database, loglike)
the_biogeme.modelName = 'b05latent_choice_full'

# Estimate the parameters
results = the_biogeme.estimate()

print(results.getBetaValues())
print(f'Estimated betas: {len(results.data.betaValues)}')
print(f'Final log likelihood: {results.data.logLike:.3f}')
print(f'Output file: {results.data.htmlFileName}')
results.writeLaTeX()
print(f'LaTeX file: {results.data.latexFileName}')



"""File b04latent_choice_seq_mc.py

Choice model with the latent variable.
Mixture of logit, with Monte-Carlo integration
Measurement equation for the indicators.
Sequential estimation.

:author: Michel Bierlaire, EPFL
:date: Thu Apr 13 18:00:05 2023

"""

import sys
import biogeme.logging as blog
import biogeme.biogeme as bio
import biogeme.exceptions as excep
from biogeme import models
import biogeme.results as res
import biogeme.messaging as msg
from biogeme.expressions import (
    Beta,
    bioDraws,
    MonteCarlo,
    exp,
    log,
)

from optima import (
    database,
    age_65_more,
    formulaIncome,
    moreThanOneCar,
    moreThanOneBike,
    individualHouse,
    male,
    haveChildren,
    haveGA,
    highEducation,
    TimePT,
    MarginalCostPT,
    TimeCar,
    CostCarCHF,
    distance_km,
    TripPurpose,
    WaitingTimePT,
    Choice,
)

logger = blog.get_screen_logger(level=blog.INFO)
logger.info('Example b04latent_choice_seq_mc.py')


# Read the estimates from the structural equation estimation
FILENAME = 'b02one_latent_ordered'
try:
    structResults = res.bioResults(pickleFile=f'{FILENAME}.pickle')
except excep.BiogemeError:
    print(
        f'Run first the script {FILENAME}.py in order to generate the '
        f'file {FILENAME}.pickle.'
    )
    sys.exit()
structBetas = structResults.getBetaValues()


# Coefficients

coef_intercept = structBetas['coef_intercept']
coef_age_65_more = structBetas['coef_age_65_more']
coef_haveGA = structBetas['coef_haveGA']
coef_moreThanOneCar = structBetas['coef_moreThanOneCar']
coef_moreThanOneBike = structBetas['coef_moreThanOneBike']
coef_individualHouse = structBetas['coef_individualHouse']
coef_male = structBetas['coef_male']
coef_haveChildren = structBetas['coef_haveChildren']
coef_highEducation = structBetas['coef_highEducation']

# Latent variable: structural equation

# Define a random parameter, normally distributed, designed to be used
# for numerical integration
sigma_s = Beta('sigma_s', 1, None, None, 0)

CARLOVERS = (
    coef_intercept
    + coef_age_65_more * age_65_more
    + formulaIncome
    + coef_moreThanOneCar * moreThanOneCar
    + coef_moreThanOneBike * moreThanOneBike
    + coef_individualHouse * individualHouse
    + coef_male * male
    + coef_haveChildren * haveChildren
    + coef_haveGA * haveGA
    + coef_highEducation * highEducation
    + sigma_s * bioDraws('EC', 'NORMAL_MLHS')
)

# Choice model
ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_PT = Beta('ASC_PT', 0, None, None, 1)
ASC_SM = Beta('ASC_SM', 0, None, None, 0)
BETA_COST_HWH = Beta('BETA_COST_HWH', 0, None, None, 0)
BETA_COST_OTHER = Beta('BETA_COST_OTHER', 0, None, None, 0)
BETA_DIST = Beta('BETA_DIST', 0, None, None, 0)
BETA_TIME_CAR_REF = Beta('BETA_TIME_CAR_REF', 0, None, 0, 0)
BETA_TIME_PT_REF = Beta('BETA_TIME_PT_REF', 0, None, 0, 0)
BETA_WAITING_TIME = Beta('BETA_WAITING_TIME', 0, None, None, 0)

# The coefficient of the latent variable should be initialized to
# something different from zero. If not, the algorithm may be trapped
# in a local optimum, and never change the value.
BETA_TIME_PT_CL = Beta('BETA_TIME_PT_CL', -0.01, None, None, 0)
BETA_TIME_CAR_CL = Beta('BETA_TIME_CAR_CL', -0.01, None, None, 0)


TimePT_scaled = database.DefineVariable('TimePT_scaled', TimePT / 200)
TimeCar_scaled = database.DefineVariable('TimeCar_scaled', TimeCar / 200)
MarginalCostPT_scaled = database.DefineVariable(
    'MarginalCostPT_scaled', MarginalCostPT / 10
)
CostCarCHF_scaled = database.DefineVariable('CostCarCHF_scaled', CostCarCHF / 10)
distance_km_scaled = database.DefineVariable('distance_km_scaled', distance_km / 5)
PurpHWH = database.DefineVariable('PurpHWH', TripPurpose == 1)
PurpOther = database.DefineVariable('PurpOther', TripPurpose != 1)

# Definition of utility functions:

BETA_TIME_PT = BETA_TIME_PT_REF * exp(BETA_TIME_PT_CL * CARLOVERS)

V0 = (
    ASC_PT
    + BETA_TIME_PT * TimePT_scaled
    + BETA_WAITING_TIME * WaitingTimePT
    + BETA_COST_HWH * MarginalCostPT_scaled * PurpHWH
    + BETA_COST_OTHER * MarginalCostPT_scaled * PurpOther
)

BETA_TIME_CAR = BETA_TIME_CAR_REF * exp(BETA_TIME_CAR_CL * CARLOVERS)

V1 = (
    ASC_CAR
    + BETA_TIME_CAR * TimeCar_scaled
    + BETA_COST_HWH * CostCarCHF_scaled * PurpHWH
    + BETA_COST_OTHER * CostCarCHF_scaled * PurpOther
)

V2 = ASC_SM + BETA_DIST * distance_km_scaled

# Associate utility functions with the numbering of alternatives
V = {0: V0, 1: V1, 2: V2}

# Conditional to omega, we have a logit model (called the kernel)
condprob = models.logit(V, None, Choice)
# We integrate over omega using numerical integration
loglike = log(MonteCarlo(condprob))

# Define level of verbosity
logger = msg.bioMessage()
# logger.setSilent()
# logger.setWarning()
logger.setGeneral()
# logger.setDetailed()

# Create the Biogeme object
the_biogeme = bio.BIOGEME(database, loglike)
the_biogeme.modelName = 'b04latent_choice_seq_mc'

# Estimate the parameters
results = the_biogeme.estimate()

print(f'Estimated betas: {len(results.data.betaValues)}')
print(f'Final log likelihood: {results.data.logLike:.3f}')
print(f'Output file: {results.data.htmlFileName}')
results.writeLaTeX()
print(f'LaTeX file: {results.data.latexFileName}')



"""Combine several arithmetic expressions and a database to obtain formulas

:author: Michel Bierlaire
:date: Sat Jul 30 12:36:40 2022
"""
import logging
from typing import NamedTuple, Dict, List
import biogeme.exceptions as excep
from biogeme.elementary_expressions import TypeOfElementaryExpression


class ElementsTuple(NamedTuple):
    expressions: Dict[str, 'Expression']
    indices: Dict[str, int]
    names: List[str]


logger = logging.getLogger(__name__)


class IdManager:
    """Class combining managing the ids of an arithmetic expression."""

    def __init__(self, expressions, database, number_of_draws):
        """Ctor

        :param expressions: list of expressions
        :type expressions: list(biogeme.expressions.Expression)

        :param database: database with the variables as column names
        :type database: biogeme.database.Database

        :param number_of_draws: number of draws for Monte-Carlo integration
        :type number_of_draws: int

        :raises BiogemeError: if an expression contains a variable and
            no database is provided.

        """
        self.expressions = expressions
        self.database = database
        self.number_of_draws = number_of_draws
        self.elementary_expressions = None
        self.free_betas = None
        self.free_betas_values = None
        self.number_of_free_betas = 0
        self.fixed_betas = None
        self.fixed_betas_values = None
        self.bounds = None
        self.random_variables = None
        self.draws = None
        self.variables = None
        self.requires_draws = False
        for f in self.expressions:
            the_variables = f.set_of_elementary_expression(
                the_type=TypeOfElementaryExpression.VARIABLE
            )
            if the_variables and database is None:
                raise excep.BiogemeError(
                    f'No database is provided and an expression '
                    f'contains variables: {the_variables}'
                )
            if f.embedExpression('MonteCarlo') or f.embedExpression('bioDraws'):
                self.requires_draws = True

        self.prepare()

    def __str__(self):
        return str(self.elementary_expressions.indices)

    def __repr__(self):
        return str(self.elementary_expressions.indices)

    def __eq__(self, other):
        return self.elementary_expressions == other.elementary_expressions

    def audit(self):
        """Performs various checks on the expressions.

        :return: tuple listOfErrors, listOfWarnings
        :rtype: list(string), list(string)
        """
        list_of_errors = []
        list_of_warnings = []
        if self.database.isPanel():
            dict_of_variables = self.expressions.dictOfVariablesOutsidePanelTrajectory()
            if dict_of_variables:
                err_msg = (
                    f'Error in the loglikelihood function. '
                    f'Some variables are not inside '
                    f'PanelLikelihoodTrajectory: '
                    f'{dict_of_variables.keys()} .'
                    f'If the database is organized as panel data, '
                    f'all variables must be used inside a '
                    f'PanelLikelihoodTrajectory. '
                    f'If it is not consistent with your model, '
                    f'generate a flat '
                    f'version of the data using the function '
                    f'`generateFlatPanelDataframe`.'
                )
                list_of_errors.append(err_msg)
        return list_of_errors, list_of_warnings

    def change_init_values(self, betas):
        """Modifies the values of the pameters

        :param betas: dictionary where the keys are the names of the
                      parameters, and the values are the new value for
                      the parameters.
        :type betas: dict(string:float)
        """

        def get_value(name):
            v = betas.get(name)
            if v is None:
                return self.free_betas.expressions[name].initValue
            return v

        self.free_betas_values = [get_value(x) for x in self.free_betas.names]

    def expressions_names_indices(self, dict_of_elements):
        """Assigns consecutive indices to expressions

        :param dict_of_elements: dictionary of expressions. The keys
            are the names.
        :type dict_of_elements: dict(str: biogeme.expressions.Expression)

        :return: a tuple with the original dictionary, the indices,
            and the sorted names.
        :rtype: ElementsTuple
        """
        indices = {}
        names = {}
        names = sorted(dict_of_elements)
        for i, v in enumerate(names):
            indices[v] = i

        return ElementsTuple(expressions=dict_of_elements, indices=indices, names=names)

    def prepare(self):
        """Extract from the formulas the literals (parameters,
        variables, random variables) and decide a numbering convention.

        The numbering is done in the following order:

        (i) free betas,
        (ii) fixed betas,
        (iii) random variables for numerical integration,
        (iv) random variables for Monte-Carlo integration,
        (v) variables

        The numbering convention will be performed for all expressions
        together, so that the same elementary expressions in several
        expressions will have the same index.


        """

        # Free parameters (to be estimated), sorted by alphabetical order
        expr = {}
        for f in self.expressions:
            d = f.dict_of_elementary_expression(
                the_type=TypeOfElementaryExpression.FREE_BETA
            )
            expr = dict(expr, **d)

        self.free_betas = self.expressions_names_indices(expr)

        self.bounds = [
            (
                self.free_betas.expressions[b].lb,
                self.free_betas.expressions[b].ub,
            )
            for b in self.free_betas.names
        ]
        self.number_of_free_betas = len(self.free_betas.names)
        # Fixed parameters (not to be estimated), sorted by alphatical order.
        expr = {}
        for f in self.expressions:
            d = f.dict_of_elementary_expression(
                the_type=TypeOfElementaryExpression.FIXED_BETA
            )
            expr = dict(expr, **d)
        self.fixed_betas = self.expressions_names_indices(expr)

        # Random variables for numerical integration
        expr = {}
        for f in self.expressions:
            d = f.dict_of_elementary_expression(
                the_type=TypeOfElementaryExpression.RANDOM_VARIABLE
            )
            expr = dict(expr, **d)
        self.random_variables = self.expressions_names_indices(expr)

        # Draws
        expr = {}
        for f in self.expressions:
            d = f.dict_of_elementary_expression(
                the_type=TypeOfElementaryExpression.DRAWS
            )
            expr = dict(expr, **d)
        self.draws = self.expressions_names_indices(expr)

        # Variables
        # Here, we do not extract the variables from the
        # formulas. Instead, we use all the variables in the database.
        if self.database is not None:
            variables_names = list(self.database.data.columns.values)
            variables_indices = {}
            for i, v in enumerate(variables_names):
                variables_indices[v] = i
            self.variables = ElementsTuple(
                expressions=None,
                indices=variables_indices,
                names=variables_names,
            )
        else:
            self.variables = ElementsTuple(expressions=None, indices=None, names=[])

        # Merge all the names
        elementary_expressions_names = (
            self.free_betas.names
            + self.fixed_betas.names
            + self.random_variables.names
            + self.draws.names
            + self.variables.names
        )

        if len(elementary_expressions_names) != len(set(elementary_expressions_names)):
            duplicates = {
                x
                for x in elementary_expressions_names
                if elementary_expressions_names.count(x) > 1
            }
            error_msg = (
                f'The following elementary expressions are defined '
                f'more than once: {duplicates}.'
            )
            raise excep.BiogemeError(error_msg)

        elementary_expressions_indices = {
            v: i for i, v in enumerate(elementary_expressions_names)
        }

        self.elementary_expressions = ElementsTuple(
            expressions=None,
            indices=elementary_expressions_indices,
            names=elementary_expressions_names,
        )

        self.free_betas_values = [
            self.free_betas.expressions[x].initValue for x in self.free_betas.names
        ]
        self.fixed_betas_values = [
            self.fixed_betas.expressions[x].initValue for x in self.fixed_betas.names
        ]

        if self.requires_draws:
            self.database.generateDraws(
                self.draws.expressions, self.draws.names, self.number_of_draws
            )

    def setDataMap(self, sample):
        """Specify the map of the panel data in the expressions

        :param sample: map of the panel data (see
            :func:`biogeme.database.Database.buildPanelMap`)
        :type sample: pandas.DataFrame
        """
        for f in self.expressions:
            f.cpp.setDataMap(sample)

    def setData(self, sample):
        """Specify the sample

        :param sample: map of the panel data (see
            :func:`biogeme.database.Database.buildPanelMap`)
        :type sample: pandas.DataFrame

        """
        for f in self.expressions:
            f.cpp.setData(sample)



"""
Test the tools module

:author: Michel Bierlaire
:date: Sun Aug  8 17:42:48 2021
"""
# Bug in pylint
# pylint: disable=no-member
#
# Too constraining
# pylint: disable=invalid-name, too-many-instance-attributes
#
# Not needed in test
# pylint: disable=missing-function-docstring, missing-class-docstring

import unittest
from copy import deepcopy
import numpy as np
import pandas as pd
from biogeme import tools
import biogeme.exceptions as excep
from test_data import (
    input_flatten,
    output_flatten_1,
    output_flatten_2,
    output_flatten_3,
)


def myFunction(x):
    f = np.log(x[0]) + np.exp(x[1])
    g = np.empty(2)
    g[0] = 1.0 / x[0]
    g[1] = np.exp(x[1])
    H = np.empty((2, 2))
    H[0, 0] = -1.0 / x[0] ** 2
    H[0, 1] = 0.0
    H[1, 0] = 0.0
    H[1, 1] = np.exp(x[1])
    return f, g, H


class TestTools(unittest.TestCase):
    def test_findiff_g(self):
        x = np.array([1.1, 1.1])
        g_fd = tools.findiff_g(myFunction, x)
        np.testing.assert_almost_equal(g_fd, [0.90909087, 3.00416619])

    def test_findiff_H(self):
        x = np.array([1.1, 1.1])
        H_fd = tools.findiff_H(myFunction, x)
        np.testing.assert_almost_equal(
            H_fd,
            [
                [
                    -0.8264462,
                    0,
                ],
                [
                    0,
                    3.00416619,
                ],
            ],
        )

    def test_checkDerivatives(self):
        x = np.array([1.1, -1.5])
        _, _, _, gdiff, hdiff = tools.checkDerivatives(
            myFunction, x, names=['First', 'Second']
        )
        np.testing.assert_almost_equal(gdiff, [0, 0])
        np.testing.assert_almost_equal(hdiff, [[0, 0], [0, 0]])

        x = np.array([3.2, 1.32])
        _, _, _, gdiff, hdiff = tools.checkDerivatives(
            myFunction, x, names=['First', 'Second']
        )
        np.testing.assert_almost_equal(gdiff, [0, 0], decimal=5)
        np.testing.assert_almost_equal(hdiff, [[0, 0], [0, 0]], decimal=5)

    def test_getPrimeNumbers(self):
        result = tools.get_prime_numbers(7)
        self.assertListEqual(result, [2, 3, 5, 7, 11, 13, 17])

        with self.assertRaises(excep.BiogemeError):
            result = tools.get_prime_numbers(0)

        with self.assertRaises(excep.BiogemeError):
            result = tools.get_prime_numbers(-1)

        with self.assertRaises(excep.BiogemeError):
            result = tools.get_prime_numbers(0.3)

    def test_calculatePrimeNumbers(self):
        result = tools.calculate_prime_numbers(10)
        self.assertListEqual(result, [2, 3, 5, 7])
        result = tools.calculate_prime_numbers(0)
        self.assertListEqual(result, [])

        with self.assertRaises(excep.BiogemeError):
            result = tools.calculate_prime_numbers(-1)

        with self.assertRaises(excep.BiogemeError):
            result = tools.calculate_prime_numbers(0.3)

    def test_countNumberOfGroups(self):
        df = pd.DataFrame(
            {
                'ID': [1, 1, 2, 3, 3, 1, 2, 3],
                'value': [1000, 2000, 3000, 4000, 5000, 5000, 10000, 20000],
            }
        )
        nbr = tools.countNumberOfGroups(df, 'ID')
        self.assertEqual(nbr, 6)
        nbr = tools.countNumberOfGroups(df, 'value')
        self.assertEqual(nbr, 7)

    def test_likelihood_ratio_test(self):
        model1 = (-1340.8, 5)
        model2 = (-1338.49, 7)
        r = tools.likelihood_ratio_test(model1, model2)
        self.assertSequenceEqual(
            r,
            (
                'H0 cannot be rejected at level 5.0%',
                4.619999999999891,
                5.991464547107979,
            ),
        )
        r = tools.likelihood_ratio_test(model2, model1)
        self.assertSequenceEqual(
            r,
            (
                'H0 cannot be rejected at level 5.0%',
                4.619999999999891,
                5.991464547107979,
            ),
        )
        r = tools.likelihood_ratio_test(model1, model2, significance_level=0.1)
        self.assertSequenceEqual(
            r,
            (
                'H0 can be rejected at level 10.0%',
                4.619999999999891,
                4.605170185988092,
            ),
        )
        with self.assertRaises(excep.BiogemeError):
            model1 = (-1340.8, 7)
            model2 = (-1338.49, 5)
            tools.likelihood_ratio_test(model1, model2)

    def test_flatten_database(self):
        df = deepcopy(input_flatten)
        result_1 = tools.flatten_database(df, 'ID', row_name='Name')
        output_flatten_1.index.name = 'ID'
        pd.testing.assert_frame_equal(result_1, output_flatten_1)
        result_2 = tools.flatten_database(input_flatten, 'ID')
        output_flatten_2.index.name = 'ID'
        pd.testing.assert_frame_equal(result_2, output_flatten_2)
        result_3 = tools.flatten_database(df, 'ID', identical_columns=[])
        output_flatten_3.index.name = 'ID'
        pd.testing.assert_frame_equal(result_3, output_flatten_3)

    def test_temporary_file(self):
        content = 'a random sentence'
        with tools.TemporaryFile() as filename:
            with open(filename, 'w', encoding='utf-8') as f:
                f.write(content)
            with open(filename, 'r', encoding='utf-8') as f:
                check = f.read()
            self.assertEqual(content, check)

    def test_model_names(self):
        model_names = tools.ModelNames(prefix='Test')

        name_a = model_names('model_a')
        correct_name_a = 'Test_000000'
        self.assertEqual(name_a, correct_name_a)
        name_b = model_names('model_b')
        correct_name_b = 'Test_000001'
        self.assertEqual(name_a, correct_name_a)
        name_a = model_names('model_a')
        self.assertEqual(name_a, correct_name_a)

    def test_generate_unique_ids(self):
        list_1 = ['A', 'B', 'C']
        correct_dict_1 = {'A': 'A', 'B': 'B', 'C': 'C'}
        dict_1 = tools.generate_unique_ids(list_1)
        self.assertDictEqual(correct_dict_1, dict_1)

        list_2 = ['A', 'B', 'B']
        correct_dict_2 = {'A': 'A', 'B_0': 'B', 'B_1': 'B'}
        dict_2 = tools.generate_unique_ids(list_2)
        self.assertDictEqual(correct_dict_2, dict_2)

        list_3 = ['B', 'B', 'B']
        correct_dict_3 = {'B_2': 'B', 'B_0': 'B', 'B_1': 'B'}
        dict_3 = tools.generate_unique_ids(list_3)
        self.assertDictEqual(correct_dict_3, dict_3)


class UniqueProductTestCase(unittest.TestCase):
    def test_unique_product(self):
        # Test case with small iterables
        iterables = [range(2), range(3), range(2)]
        expected_output = [
            (0, 0, 0),
            (0, 0, 1),
            (0, 1, 0),
            (0, 1, 1),
            (0, 2, 0),
            (0, 2, 1),
            (1, 0, 0),
            (1, 0, 1),
            (1, 1, 0),
            (1, 1, 1),
            (1, 2, 0),
            (1, 2, 1),
        ]
        result = list(tools.unique_product(*iterables))
        self.assertEqual(result, expected_output)

        # Test case with larger iterables
        iterables = [range(3), range(4), range(2)]
        expected_output = [
            (0, 0, 0),
            (0, 0, 1),
            (0, 1, 0),
            (0, 1, 1),
            (0, 2, 0),
            (0, 2, 1),
            (0, 3, 0),
            (0, 3, 1),
            (1, 0, 0),
            (1, 0, 1),
            (1, 1, 0),
            (1, 1, 1),
            (1, 2, 0),
            (1, 2, 1),
            (1, 3, 0),
            (1, 3, 1),
            (2, 0, 0),
            (2, 0, 1),
            (2, 1, 0),
            (2, 1, 1),
            (2, 2, 0),
            (2, 2, 1),
            (2, 3, 0),
            (2, 3, 1),
        ]
        result = list(tools.unique_product(*iterables))
        self.assertEqual(result, expected_output)

        # Test case with empty iterable
        iterables = [range(3), [], range(2)]
        expected_output = []
        result = list(tools.unique_product(*iterables))
        self.assertEqual(result, expected_output)

        # Test case with single iterable
        iterables = [range(3)]
        expected_output = [(0,), (1,), (2,)]
        result = list(tools.unique_product(*iterables))
        self.assertEqual(result, expected_output)


if __name__ == '__main__':
    unittest.main()



"""File: b02simple_integral.py

 Author: Michel Bierlaire, EPFL
 Date: Thu Apr 13 20:46:01 2023

Calculation of a simple integral using numerical integration and
Monte-Carlo integration with various types of draws, including Halton
draws base 13. It illustrates how to use draws that are not directly
available in Biogeme.

"""


import pandas as pd
import biogeme.database as db
import biogeme.biogeme as bio
from biogeme import draws
from biogeme.expressions import exp, bioDraws, MonteCarlo


# We create a fake database with one entry, as it is required
# to store the draws
pandas = pd.DataFrame()
pandas['FakeColumn'] = [1.0]
database = db.Database('fakeDatabase', pandas)


def halton13(sample_size, number_of_draws):
    """
    The user can define new draws. For example, Halton draws
    with base 13, skipping the first 10 draws.
    """
    return draws.getHaltonDraws(sample_size, number_of_draws, base=13, skip=10)


mydraws = {'HALTON13': (halton13, 'Halton draws, base 13, skipping 10')}
database.setRandomNumberGenerators(mydraws)

integrand = exp(bioDraws('U', 'UNIFORM'))
simulatedI = MonteCarlo(integrand)

integrand_halton = exp(bioDraws('U_halton', 'UNIFORM_HALTON2'))
simulatedI_halton = MonteCarlo(integrand_halton)

integrand_halton13 = exp(bioDraws('U_halton13', 'HALTON13'))
simulatedI_halton13 = MonteCarlo(integrand_halton13)

integrand_mlhs = exp(bioDraws('U_mlhs', 'UNIFORM_MLHS'))
simulatedI_mlhs = MonteCarlo(integrand_mlhs)

trueI = exp(1.0) - 1.0

R = 20000

sampleVariance = MonteCarlo(integrand * integrand) - simulatedI * simulatedI
stderr = (sampleVariance / R) ** 0.5
error = simulatedI - trueI

sampleVariance_halton = (
    MonteCarlo(integrand_halton * integrand_halton)
    - simulatedI_halton * simulatedI_halton
)
stderr_halton = (sampleVariance_halton / R) ** 0.5
error_halton = simulatedI_halton - trueI

sampleVariance_halton13 = (
    MonteCarlo(integrand_halton13 * integrand_halton13)
    - simulatedI_halton13 * simulatedI_halton13
)
stderr_halton13 = (sampleVariance_halton13 / R) ** 0.5
error_halton13 = simulatedI_halton13 - trueI

sampleVariance_mlhs = (
    MonteCarlo(integrand_mlhs * integrand_mlhs) - simulatedI_mlhs * simulatedI_mlhs
)
stderr_mlhs = (sampleVariance_mlhs / R) ** 0.5
error_mlhs = simulatedI_mlhs - trueI


simulate = {
    'Analytical Integral': trueI,
    'Simulated Integral': simulatedI,
    'Sample variance   ': sampleVariance,
    'Std Error         ': stderr,
    'Error             ': error,
    'Simulated Integral (Halton)': simulatedI_halton,
    'Sample variance (Halton)   ': sampleVariance_halton,
    'Std Error (Halton)         ': stderr_halton,
    'Error (Halton)             ': error_halton,
    'Simulated Integral (Halton13)': simulatedI_halton13,
    'Sample variance (Halton13)   ': sampleVariance_halton13,
    'Std Error (Halton13)         ': stderr_halton13,
    'Error (Halton13)             ': error_halton13,
    'Simulated Integral (MLHS)': simulatedI_mlhs,
    'Sample variance (MLHS)   ': sampleVariance_mlhs,
    'Std Error (MLHS)         ': stderr_mlhs,
    'Error (MLHS)             ': error_mlhs,
}

biosim = bio.BIOGEME(database, simulate)
biosim.modelName = 'b02simple_integral'
results = biosim.simulate(theBetaValues={})
print(f'Analytical integral:{results.iloc[0]["Analytical Integral"]:.6g}')
print('\t\tUniform\t\tHalton\t\tHalton13\tMLHS')
print(
    f'Simulated\t{results.iloc[0]["Simulated Integral"]:.6g}\t'
    f'{results.iloc[0]["Simulated Integral (Halton)"]:.6g}\t'
    f'{results.iloc[0]["Simulated Integral (Halton13)"]:.6g}\t'
    f'{results.iloc[0]["Simulated Integral (MLHS)"]:.6g}'
)
print(
    f'Sample var.\t{results.iloc[0]["Sample variance   "]:.6g}\t'
    f'{results.iloc[0]["Sample variance (Halton)   "]:.6g}\t'
    f'{results.iloc[0]["Sample variance (Halton13)   "]:.6g}\t'
    f'{results.iloc[0]["Sample variance (MLHS)   "]:.6g}'
)
print(
    f'Std error\t{results.iloc[0]["Std Error         "]:.6g}\t'
    f'{results.iloc[0]["Std Error (Halton)         "]:.6g}\t'
    f'{results.iloc[0]["Std Error (Halton13)         "]:.6g}\t'
    f'{results.iloc[0]["Std Error (MLHS)         "]:.6g}'
)
print(
    f'Error\t\t{results.iloc[0]["Error             "]:.6g}\t'
    f'{results.iloc[0]["Error (Halton)             "]:.6g}\t'
    f'{results.iloc[0]["Error (Halton13)             "]:.6g}\t'
    f'{results.iloc[0]["Error (MLHS)             "]:.6g}'
)



"""File b01logit_simul.py

:author: Michel Bierlaire, EPFL
:date: Sun Apr  9 17:13:23 2023

 Example of simulation with a logit model
 Three alternatives: Train, Car and Swissmetro
 SP data
"""

import biogeme.biogeme as bio
from biogeme import models
from biogeme.expressions import Beta, Derive
from swissmetro_data import (
    database,
    SM_AV,
    CAR_AV_SP,
    TRAIN_AV_SP,
    TRAIN_TT,
    TRAIN_TT_SCALED,
    TRAIN_COST_SCALED,
    SM_TT,
    SM_TT_SCALED,
    SM_COST_SCALED,
    CAR_TT,
    CAR_TT_SCALED,
    CAR_CO_SCALED,
)

# Parameters to be estimated
ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_TRAIN = Beta('ASC_TRAIN', 0, None, None, 0)
ASC_SM = Beta('ASC_SM', 0, None, None, 1)
B_TIME = Beta('B_TIME', 0, None, None, 0)
B_COST = Beta('B_COST', 0, None, None, 0)

# Definition of the utility functions
V1 = ASC_TRAIN + B_TIME * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED
V2 = ASC_SM + B_TIME * SM_TT_SCALED + B_COST * SM_COST_SCALED
V3 = ASC_CAR + B_TIME * CAR_TT_SCALED + B_COST * CAR_CO_SCALED

# Associate utility functions with the numbering of alternatives
V = {1: V1, 2: V2, 3: V3}

# Associate the availability conditions with the alternatives
av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

# The choice model is a logit, with availability conditions
prob1 = models.logit(V, av, 1)
prob2 = models.logit(V, av, 2)
prob3 = models.logit(V, av, 3)


# Elasticities can be computed. We illustrate below two
# formulas. Check in the output file that they produce the same
# result.

# First, the general definition of elasticities. This illustrates the
# use of the Derive expression, and can be used with any model,
# however complicated it is. Note the quotes in the Derive opertor.
genelas1 = Derive(prob1, 'TRAIN_TT') * TRAIN_TT / prob1
genelas2 = Derive(prob2, 'SM_TT') * SM_TT / prob2
genelas3 = Derive(prob3, 'CAR_TT') * CAR_TT / prob3

# Second, the elasticity of logit models. See Ben-Akiva and Lerman for
# the formula

logitelas1 = TRAIN_AV_SP * (1.0 - prob1) * TRAIN_TT_SCALED * B_TIME
logitelas2 = SM_AV * (1.0 - prob2) * SM_TT_SCALED * B_TIME
logitelas3 = CAR_AV_SP * (1.0 - prob3) * CAR_TT_SCALED * B_TIME

simulate = {
    'Prob. train': prob1,
    'Prob. Swissmetro': prob2,
    'Prob. car': prob3,
    'logit elas. 1': logitelas1,
    'generic elas. 1': genelas1,
    'logit elas. 2': logitelas2,
    'generic elas. 2': genelas2,
    'logit elas. 3': logitelas3,
    'generic elas. 3': genelas3,
}


biosim = bio.BIOGEME(database, simulate)
biosim.modelName = 'b01logit_simul'
betas = {
    'ASC_TRAIN': -0.701188,
    'B_TIME': -1.27786,
    'B_COST': -1.08379,
    'ASC_SM': 0,
    'ASC_CAR': -0.154633,
}


results = biosim.simulate(theBetaValues=betas)
print(results.describe())



""" Defines the types of the elementary expressions

:author: Michel Bierlaire
:date: Tue Mar  7 18:38:21 2023

"""
from enum import Enum, auto


class TypeOfElementaryExpression(Enum):
    """
    Defines the types of elementary expressions
    """

    VARIABLE = auto()
    BETA = auto()
    FREE_BETA = auto()
    FIXED_BETA = auto()
    RANDOM_VARIABLE = auto()
    DRAWS = auto()



""" Singleton class

:author: Michel Bierlaire
:date: Sat Jul 18 16:45:35 2020

"""


class Singleton(type):
    """
    A singleton is a class with only one instance
    """

    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]



import os
import unittest
import pandas as pd
import biogeme.database as db
import biogeme.biogeme as bio
import biogeme.distributions as dist
from biogeme.expressions import Beta, log, Elem


myPath = os.path.dirname(os.path.abspath(__file__))
df = pd.read_csv(f'{myPath}/swissmetro.dat', sep='\t')
database = db.Database('swissmetro', df)

# The Pandas data structure is available as database.data. Use all the
# Pandas functions to invesigate the database
# print(database.data.describe())

globals().update(database.variables)

exclude = ((PURPOSE != 1) * (PURPOSE != 3) + (CHOICE == 0)) > 0
database.remove(exclude)


B_TIME = Beta('B_TIME', 0, None, None, 0)
B_COST = Beta('B_COST', 0, None, None, 0)

tau1 = Beta('tau1', -1, None, 0, 0)
delta2 = Beta('delta2', 2, 0, None, 0)

tau2 = tau1 + delta2


TRAIN_COST = TRAIN_CO * (GA == 0)

TRAIN_TT_SCALED = database.DefineVariable('TRAIN_TT_SCALED', TRAIN_TT / 100.0)
TRAIN_COST_SCALED = database.DefineVariable(
    'TRAIN_COST_SCALED', TRAIN_COST / 100
)

#  Utility

U = B_TIME * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED


ChoiceProba = {
    1: 1 - dist.logisticcdf(U - tau1),
    2: dist.logisticcdf(U - tau1) - dist.logisticcdf(U - tau2),
    3: dist.logisticcdf(U - tau2),
}

logprob = log(Elem(ChoiceProba, CHOICE))


class test_18(unittest.TestCase):
    def testEstimation(self):
        biogeme = bio.BIOGEME(database, logprob)
        biogeme.saveIterations = False
        biogeme.generateHtml = False
        biogeme.generatePickle = False
        results = biogeme.estimate()
        self.assertAlmostEqual(results.data.logLike, -5789.309, 2)


if __name__ == '__main__':
    unittest.main()



"""
Test the version module

:author: Michel Bierlaire
:date: Fri Feb 24 07:57:32 2023

"""
import unittest
import biogeme.version as ver


class TestVersion(unittest.TestCase):
    def test_get_html(self):
        html = ver.getHtml()
        self.assertIsInstance(html, str, 'Not a string')
        text = ver.getText()
        self.assertIsInstance(text, str, 'Not a string')
        latex = ver.getLaTeX()
        self.assertIsInstance(latex, str, 'Not a string')



"""Defines a generic exception for Biogeme

:author: Michel Bierlaire

:date: Tue Mar 26 16:47:11 2019

"""

# Too constraining
# pylint: disable=invalid-name


class BiogemeError(Exception):
    """Defines a generic exception for Biogeme."""


class ValueOutOfRange(BiogemeError):
    """Value of out range error."""


class DuplicateError(BiogemeError):
    """duplicate error."""



"""
Optimization algorithms for Biogeme

:author: Michel Bierlaire
:date: Mon Dec 21 10:24:24 2020

"""

# There seems to be a bug in PyLint.
# pylint: disable=invalid-unary-operand-type, no-member

# Too constraining
# pylint: disable=invalid-name
# pylint: disable=too-many-lines, too-many-locals
# pylint: disable=too-many-arguments, too-many-branches
# pylint: disable=too-many-statements, too-many-return-statements
# pylint: disable=bare-except


import logging
import numpy as np
import scipy.optimize as sc
from biogeme_optimization.bounds import Bounds
from biogeme_optimization.linesearch import newton_linesearch, bfgs_linesearch
from biogeme_optimization.trust_region import newton_trust_region, bfgs_trust_region
from biogeme_optimization.simple_bounds import simple_bounds_newton_algorithm
import biogeme.exceptions as excep

logger = logging.getLogger(__name__)


def scipy(fct, initBetas, bounds, variable_names, parameters=None):
    """Optimization interface for Biogeme, based on the scipy
    minimize function.

    :param fct: object to calculate the objective function and its derivatives.
    :type fct: algorithms.functionToMinimize

    :param initBetas: initial value of the beta parameters
    :type initBetas: numpy.array

    :param bounds: list of tuples (ell,u) containing the lower and upper bounds
          for each free parameter
    :type bounds: list(tuple)

    :param variable_names: names of the variables. Ignored
        here. Included to comply with the syntax.
    :type variable_names: list(str)

    :param parameters: dict of parameters to be transmitted to the
         optimization routine. See the `scipy`_ documentation.

    .. _`scipy`: https://docs.scipy.org/doc/scipy/reference/optimize.html

    :type parameters: dict(string:float or string)

    :return: x, messages

        - x is the solution generated by the algorithm,
        - messages is a dictionary describing several information
          about the algorithm

    :rtype: numpay.array, dict(str:object)

    """

    def f_and_grad(x):
        fct.set_variables(x)
        function_data = fct.f_g()
        return function_data.function, function_data.gradient

    logger.info('Optimization algorithm: scipy')
    # Absolute tolerance
    absgtol = 1.0e-7
    opts = {'ftol': np.finfo(np.float64).eps, 'gtol': absgtol}
    if parameters is not None:
        opts = {**opts, **parameters}

    if 'gtol' in opts.keys():
        logger.info(f'Minimize with tol {opts["gtol"]}')

    results = sc.minimize(f_and_grad, initBetas, bounds=bounds, jac=True, options=opts)

    messages = {
        'Algorithm': 'scipy.optimize',
        'Cause of termination': results.message,
        'Number of iterations': results.nit,
        'Number of function evaluations': results.nfev,
    }

    return results.x, messages


def newton_linesearch_for_biogeme(
    fct, initBetas, bounds, variable_names, parameters=None
):
    """Optimization interface for Biogeme, based on Newton method.

    :param fct: object to calculate the objective function and its derivatives.
    :type fct: algorithms.functionToMinimize

    :param initBetas: initial value of the parameters.
    :type initBetas: numpy.array

    :param bounds: list of tuples (ell,u) containing the lower and
                   upper bounds for each free parameter. Note that
                   this algorithm does not support bound constraints.
                   Therefore, all the bounds must be None.

    :type bounds: list(tuples)

    :param variable_names: names of the variables.
    :type variable_names: list(str)

    :param parameters: dict of parameters to be transmitted to the
        optimization routine:

        - tolerance: when the relative gradient is below that
          threshold, the algorithm has reached convergence
          (default: :math:`\\varepsilon^{\\frac{1}{3}}`);
        - maxiter: the maximum number of iterations (default: 100).

    :type parameters: dict(string:float or int)

    :return: tuple x, nit, nfev, message, where

            - x is the solution found,
            - messages is a dictionary reporting various aspects
              related to the run of the algorithm.

    :rtype: numpy.array, dict(str:object)

    :raises biogeme.exceptions.BiogemeError: if bounds are imposed on
                the variables.

    """
    logger.info('Optimization algorithm: Newton with line search [LS-newton]')

    for ell, u in bounds:
        if ell is not None or u is not None:
            errorMsg = (
                'This algorithm does not handle bound constraints. '
                'Remove the bounds, or select another algorithm.'
            )
            raise excep.BiogemeError(errorMsg)

    maxiter = 100
    if parameters is not None:
        if 'maxiter' in parameters:
            maxiter = parameters['maxiter']

    logger.info('** Optimization: Newton with linesearch')
    return newton_linesearch(
        the_function=fct, starting_point=initBetas, maxiter=maxiter
    )


def newton_trust_region_for_biogeme(
    fct, initBetas, bounds, variable_names, parameters=None
):
    """Optimization interface for Biogeme, based on Newton method with TR.

    :param fct: object to calculate the objective function and its derivatives.
    :type fct: algorithms.functionToMinimize

    :param initBetas: initial value of the parameters.
    :type initBetas: numpy.array

    :param bounds: list of tuples (ell, u) containing the lower and
                   upper bounds for each free parameter. Note that
                   this algorithm does not support bound constraints.
                   Therefore, all the bounds must be None.
    :type bounds: list(tuples)

    :param variable_names: names of the variables.
    :type variable_names: list(str)

    :param parameters: dict of parameters to be transmitted to the
        optimization routine:

        - tolerance: when the relative gradient is below that threshold,
          the algorithm has reached convergence
          (default:  :math:`\\varepsilon^{\\frac{1}{3}}`);
        - maxiter: the maximum number of iterations (default: 100).
        - dogleg: if True, the trust region subproblem is solved using
          the Dogleg method. If False, it is solved using the
          truncated conjugate gradient method (default: False).
        - radius: the initial radius of the truat region (default: 1.0).

    :type parameters: dict(string:float or int)

    :return: tuple x, messages, where

            - x is the solution found,
            - messages is a dictionary reporting various aspects
              related to the run of the algorithm.

    :rtype: numpy.array, dict(str:object)

    :raises biogeme.exceptions.BiogemeError: if bounds are imposed on
          the variables.

    """
    logger.info('Optimization algorithm: Newton with trust region [TR-newton]')

    for ell, u in bounds:
        if ell is not None or u is not None:
            errorMsg = (
                'This algorithm does not handle bound constraints. '
                'Remove the bounds, or select another algorithm.'
            )
            raise excep.BiogemeError(errorMsg)

    maxiter = 100
    apply_dogleg = False
    radius = 1.0
    if parameters is not None:
        if 'maxiter' in parameters:
            maxiter = parameters['maxiter']
        if 'dogleg' in parameters:
            apply_dogleg = parameters['dogleg']
        if 'radius' in parameters:
            radius = parameters['radius']

    logger.info('** Optimization: Newton with trust region')
    return newton_trust_region(
        the_function=fct,
        starting_point=initBetas,
        use_dogleg=apply_dogleg,
        maxiter=maxiter,
        initial_radius=radius,
    )


def bfgs_linesearch_for_biogeme(
    fct, initBetas, bounds, variable_names, parameters=None
):
    """Optimization interface for Biogeme, based on BFGS
    quasi-Newton method with LS.

    :param fct: object to calculate the objective function and its derivatives.

    :type fct: algorithms.functionToMinimize

    :param initBetas: initial value of the parameters.

    :type initBetas: numpy.array

    :param bounds: list of tuples (ell,u) containing the lower and
                   upper bounds for each free parameter. Note that
                   this algorithm does not support bound constraints.
                   Therefore, all the bounds must be None.

    :type bounds: list(tuples)

    :param variable_names: names of the variables.
    :type variable_names: list(str)

    :param parameters: dict of parameters to be transmitted to the
        optimization routine:

        - tolerance: when the relative gradient is below that
          threshold, the algorithm has reached convergence
          (default: :math:`\\varepsilon^{\\frac{1}{3}}`);

        - maxiter: the maximum number of iterations (default: 100).

        - | initBfgs: the positive definite matrix that initalizes the
                      BFGS updates. If None, the identity matrix is
                      used. Default: None.


    :type parameters: dict(string:float or int)

    :return: tuple x, messages, where

            - x is the solution found,

            - messages is a dictionary reporting various aspects
              related to the run of the algorithm.

    :rtype: numpy.array, dict(str:object)

    :raises biogeme.exceptions.BiogemeError: if bounds are imposed on
                                             the variables.

    """
    logger.info('Optimization algorithm: BFGS with line search [LS-BFGS]')

    for ell, u in bounds:
        if ell is not None or u is not None:
            errorMsg = (
                'This algorithm does not handle bound constraints. '
                'Remove the bounds, or select another algorithm.'
            )
            raise excep.BiogemeError(errorMsg)

    maxiter = 100
    init_bfgs = None
    if parameters is not None:
        if 'maxiter' in parameters:
            maxiter = parameters['maxiter']
        if 'initBfgs' in parameters:
            init_bfgs = parameters['initBfgs']

    logger.info('** Optimization: BFGS with line search')
    return bfgs_linesearch(
        the_function=fct, starting_point=initBetas, init_bfgs=init_bfgs, maxiter=maxiter
    )


def bfgs_trust_region_for_biogeme(
    fct, initBetas, bounds, variable_names, parameters=None
):
    """Optimization interface for Biogeme, based on Newton method with TR.

    :param fct: object to calculate the objective function and its derivatives.
    :type fct: algorithms.functionToMinimize

    :param initBetas: initial value of the parameters.
    :type initBetas: numpy.array

    :param bounds: list of tuples (ell,u) containing the lower and
                   upper bounds for each free parameter. Note that
                   this algorithm does not support bound constraints.
                   Therefore, all the bounds must be None.
    :type bounds: list(tuples)

    :param variable_names: names of the variables.
    :type variable_names: list(str)

    :param parameters: dict of parameters to be transmitted to the
         optimization routine:

         - tolerance: when the relative gradient is below that
           threshold, the algorithm has reached convergence
           (default: :math:`\\varepsilon^{\\frac{1}{3}}`);

         - maxiter: the maximum number of iterations (default: 100).

         - dogleg: if True, the trust region subproblem is solved using
           the Dogleg method. If False, it is solved using the
           truncated conjugate gradient method (default: False).

         - radius: the initial radius of the truat region (default: 1.0).

         - initBfgs: the positive definite matrix that initalizes the
           BFGS updates. If None, the identity matrix is
           used. Default: None.


    :type parameters: dict(string:float or int)

    :return: tuple x, messages, where

            - x is the solution found,
            - messages is a dictionary reporting various aspects
              related to the run of the algorithm.
    :rtype: numpy.array, dict(str:object)

    :raises biogeme.exceptions.BiogemeError: if bounds are imposed on
            the variables.

    """
    logger.info('Optimization algorithm: BFGS with trust region [TR-BFGS]')
    for ell, u in bounds:
        if ell is not None or u is not None:
            errorMsg = (
                'This algorithm does not handle bound constraints. '
                'Remove the bounds, or select another algorithm.'
            )
            raise excep.BiogemeError(errorMsg)

    maxiter = 100
    apply_dogleg = False
    radius = 1.0
    init_bfgs = None
    if parameters is not None:
        if 'maxiter' in parameters:
            maxiter = parameters['maxiter']
        if 'dogleg' in parameters:
            apply_dogleg = parameters['dogleg']
        if 'radius' in parameters:
            radius = parameters['radius']
        if 'initBfgs' in parameters:
            init_bfgs = parameters['initBfgs']

    logger.info('** Optimization: BFGS with trust region')
    return bfgs_trust_region(
        the_function=fct,
        starting_point=initBetas,
        init_bfgs=init_bfgs,
        use_dogleg=apply_dogleg,
        maxiter=maxiter,
        initial_radius=radius,
    )


def simple_bounds_newton_algorithm_for_biogeme(
    fct, initBetas, bounds, variable_names, parameters=None
):
    """Optimization interface for Biogeme, based on variants of Newton
    method with simple bounds.

    :param fct: object to calculate the objective function and its derivatives.
    :type fct: algorithms.functionToMinimize

    :param initBetas: initial value of the parameters.
    :type initBetas: numpy.array

    :param bounds: list of tuples (ell,u) containing the lower and upper
                   bounds for each free
                   parameter. Note that this algorithm does not support bound
                   constraints.
                   Therefore, all the bounds must be None.
    :type bounds: list(tuples)

    :param variable_names: names of the variables.
    :type variable_names: list(str)

    :param parameters: dict of parameters to be transmitted to the
        optimization routine:

        - tolerance: when the relative gradient is below that threshold,
          the algorithm has reached convergence
          (default:  :math:`\\varepsilon^{\\frac{1}{3}}`);
        - steptol: the algorithm stops when the relative change in x
          is below this threshold. Basically, if p significant digits
          of x are needed, steptol should be set to 1.0e-p. Default:
          :math:`10^{-5}`
        - cgtolerance: when the norm of the residual is below that
          threshold, the conjugate gradient algorithm has reached
          convergence (default:  :math:`\\varepsilon^{\\frac{1}{3}}`);
        - proportionAnalyticalHessian: proportion (between 0 and 1) of
          iterations when the analytical Hessian is calculated (default: 1).
        - infeasibleConjugateGradient: if True, the conjugate gradient
          algorithm may generate infeasible solutiona until
          termination.  The result will then be projected on the
          feasible domain.  If False, the algorithm stops as soon as
          an infeasible iterate is generated (default: False).
        - maxiter: the maximum number of iterations (default: 1000).
        - radius: the initial radius of the trust region (default: 1.0).
        - eta1: threshold for failed iterations (default: 0.01).
        - eta2: threshold for very successful iteration (default 0.9).
        - enlargingFactor: factor used to enlarge the trust region
          during very successful iterations (default 10).

    :type parameters: dict(string:float or int)

    :return: x, messages

        - x is the solution generated by the algorithm,
        - messages is a dictionary describing information about the lagorithm

    :rtype: numpay.array, dict(str:object)

    """
    logger.info(
        'Optimization algorithm: hybrid Newton/BFGS with simple bounds [simple_bounds]'
    )

    cgtol = np.finfo(np.float64).eps ** 0.3333
    maxiter = 1000
    radius = 1.0
    eta1 = 0.1
    eta2 = 0.9
    proportionTrueHessian = 1.0
    enlargingFactor = 2

    # We replace the default value by user defined value, if any.
    if parameters is not None:
        if 'cgtolerance' in parameters:
            cgtol = parameters['cgtolerance']
        if 'maxiter' in parameters:
            maxiter = parameters['maxiter']
        if 'radius' in parameters:
            radius = parameters['radius']
        if 'eta1' in parameters:
            eta1 = parameters['eta1']
        if 'eta2' in parameters:
            eta2 = parameters['eta2']
        if 'enlargingFactor' in parameters:
            enlargingFactor = parameters['enlargingFactor']
        if 'proportionAnalyticalHessian' in parameters:
            proportionTrueHessian = parameters['proportionAnalyticalHessian']

    if proportionTrueHessian == 1.0:
        logger.info('** Optimization: Newton with trust region for simple bounds')
    elif proportionTrueHessian == 0.0:
        logger.info('** Optimization: BFGS with trust region for simple bounds')
    else:
        logger.info(
            f'** Optimization: Hybrid Newton '
            f'{100*proportionTrueHessian}%/BFGS '
            f'with trust region for simple bounds'
        )
    return simple_bounds_newton_algorithm(
        the_function=fct,
        bounds=Bounds(bounds),
        starting_point=initBetas,
        variable_names=variable_names,
        proportion_analytical_hessian=proportionTrueHessian,
        first_radius=radius,
        cgtol=cgtol,
        maxiter=maxiter,
        eta1=eta1,
        eta2=eta2,
        enlarging_factor=enlargingFactor,
    )


def bio_newton(fct, initBetas, bounds, variable_names, parameters=None):
    """Optimization interface for Biogeme, based on Newton's method with simple
    bounds.

    :param fct: object to calculate the objective function and its derivatives.
    :type fct: algorithms.functionToMinimize

    :param initBetas: initial value of the parameters.
    :type initBetas: numpy.array

    :param bounds: list of tuples (ell,u) containing the lower and upper
                   bounds for each free
                   parameter. Note that this algorithm does not support bound
                   constraints.
                   Therefore, all the bounds must be None.
    :type bounds: list(tuples)

    :param variable_names: names of the variables.
    :type variable_names: list(str)

    :param parameters: dict of parameters to be transmitted to the
        optimization routine:

        - tolerance: when the relative gradient is below that threshold,
          the algorithm has reached convergence
          (default:  :math:`\\varepsilon^{\\frac{1}{3}}`);
        - steptol: the algorithm stops when the relative change in x
          is below this threshold. Basically, if p significant digits
          of x are needed, steptol should be set to 1.0e-p. Default:
          :math:`10^{-5}`
        - cgtolerance: when the norm of the residual is below that
          threshold, the conjugate gradient algorithm has reached
          convergence (default:  :math:`\\varepsilon^{\\frac{1}{3}}`);
        - infeasibleConjugateGradient: if True, the conjugate
          gradient algorithm may generate until termination.  The
          result will then be projected on the feasible domain.  If
          False, the algorithm stops as soon as an infeasible iterate
          is generated (default: False).
        - maxiter: the maximum number of iterations (default: 1000).
        - radius: the initial radius of the truat region (default: 1.0).
        - eta1: threshold for failed iterations (default: 0.01).
        - eta2: threshold for very successful iteration (default 0.9).
        - enlargingFactor: factor used to enlarge the trust region
          during very successful iterations (default 10).

    :type parameters: dict(string:float or int)

    :return: x, messages

        - x is the solution generated by the algorithm,
        - messages is a dictionary describing information about the lagorithm

    :rtype: numpay.array, dict(str:object)

    """
    logger.info(
        'Optimization algorithm: Newton with simple bounds [simple_bounds_newton].'
    )

    if parameters is None:
        parameters = {'proportionAnalyticalHessian': 1}
    else:
        parameters['proportionAnalyticalHessian'] = 1
    return simple_bounds_newton_algorithm_for_biogeme(
        fct, initBetas, bounds, variable_names, parameters
    )


def bio_bfgs(fct, initBetas, bounds, variable_names, parameters=None):
    """Optimization interface for Biogeme, based on BFGS quasi-Newton
    method with simple bounds.

    :param fct: object to calculate the objective function and its derivatives.
    :type fct: algorithms.functionToMinimize

    :param initBetas: initial value of the parameters.
    :type initBetas: numpy.array

    :param bounds: list of tuples (ell,u) containing the lower and upper
                   bounds for each free
                   parameter. Note that this algorithm does not support bound
                   constraints.
                   Therefore, all the bounds must be None.
    :type bounds: list(tuples)

    :param variable_names: names of the variables.
    :type variable_names: list(str)

    :param parameters: dict of parameters to be transmitted to the
        optimization routine:

        - tolerance: when the relative gradient is below that threshold,
          the algorithm has reached convergence
          (default:  :math:`\\varepsilon^{\\frac{1}{3}}`);
        - steptol: the algorithm stops when the relative change in x
          is below this threshold. Basically, if p significant digits
          of x are needed, steptol should be set to 1.0e-p. Default:
          :math:`10^{-5}`
        - cgtolerance: when the norm of the residual is below that
          threshold, the conjugate gradient algorithm has reached
          convergence (default:  :math:`\\varepsilon^{\\frac{1}{3}}`);
        - infeasibleConjugateGradient: if True, the conjugate
          gradient algorithm may generate until termination.  The
          result will then be projected on the feasible domain.  If
          False, the algorithm stops as soon as an infeasible iterate
          is generated (default: False).
        - maxiter: the maximum number of iterations (default: 1000).
        - radius: the initial radius of the truat region (default: 1.0).
        - eta1: threshold for failed iterations (default: 0.01).
        - eta2: threshold for very successful iteration (default 0.9).
        - enlargingFactor: factor used to enlarge the trust region
          during very successful iterations (default 10).

    :type parameters: dict(string:float or int)

    :return: x, messages

        - x is the solution generated by the algorithm,
        - messages is a dictionary describing information about the algorithm

    :rtype: numpay.array, dict(str:object)

    """
    logger.info('Optimization algorithm: BFGS with simple bounds [simple_bounds_BFGS].')

    if parameters is None:
        parameters = {'proportionAnalyticalHessian': 0}
    else:
        parameters['proportionAnalyticalHessian'] = 0
    return simple_bounds_newton_algorithm_for_biogeme(
        fct, initBetas, bounds, variable_names, parameters
    )


algorithms = {
    'scipy': scipy,
    'LS-newton': newton_linesearch_for_biogeme,
    'TR-newton': newton_trust_region_for_biogeme,
    'LS-BFGS': bfgs_linesearch_for_biogeme,
    'TR-BFGS': bfgs_trust_region_for_biogeme,
    'simple_bounds': simple_bounds_newton_algorithm_for_biogeme,
    'simple_bounds_newton': bio_newton,
    'simple_bounds_BFGS': bio_bfgs,
}



""" File: b04normal_mixture_numerical.py

 Author: Michel Bierlaire, EPFL
 Date: Thu Apr 13 20:51:32 2023

Calculation of a mixtures of logit models where the integral is
calculated using numerical integration.
"""

import biogeme.biogeme as bio
import biogeme.distributions as dist
from biogeme.expressions import RandomVariable, Integrate
from biogeme import models

from swissmetro_one import (
    database,
    TRAIN_TT_SCALED,
    TRAIN_COST_SCALED,
    SM_TT_SCALED,
    SM_COST_SCALED,
    CAR_TT_SCALED,
    CAR_CO_SCALED,
    TRAIN_AV_SP,
    SM_AV,
    CAR_AV_SP,
    CHOICE,
)

# Parameters
ASC_CAR = 0.137
ASC_TRAIN = -0.402
ASC_SM = 0
B_TIME = -2.26
B_TIME_S = 1.66
B_COST = -1.29

# Define a random parameter, normally distributed,
# designed to be used for integration
omega = RandomVariable('omega')
density = dist.normalpdf(omega)
B_TIME_RND = B_TIME + B_TIME_S * omega


# Definition of the utility functions
V1 = ASC_TRAIN + B_TIME_RND * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED
V2 = ASC_SM + B_TIME_RND * SM_TT_SCALED + B_COST * SM_COST_SCALED
V3 = ASC_CAR + B_TIME_RND * CAR_TT_SCALED + B_COST * CAR_CO_SCALED

# Associate utility functions with the numbering of alternatives
V = {1: V1, 2: V2, 3: V3}

# Associate the availability conditions with the alternatives
av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

# The choice model is a logit, with availability conditions
integrand = models.logit(V, av, CHOICE)
numericalI = Integrate(integrand * density, 'omega')

simulate = {'Numerical': numericalI}

biosim = bio.BIOGEME(database, simulate)
results = biosim.simulate(theBetaValues={})
print('Mixture of logit - numerical integration: ', results.iloc[0]['Numerical'])



"""Defines the interface for a catalog of expressions that may be
considered in a specification

:author: Michel Bierlaire
:date: Sun Feb  5 15:34:56 2023

"""
import logging
import abc
from typing import NamedTuple
import biogeme.expressions as ex
import biogeme.exceptions as excep
from biogeme.configuration import (
    SEPARATOR,
    SELECTION_SEPARATOR,
)

logger = logging.getLogger(__name__)


class NamedExpression(NamedTuple):
    name: str
    expression: ex.Expression


class CatalogItem(NamedTuple):
    catalog_name: str
    item_index: int
    item_name: str


class MultipleExpression(ex.Expression, metaclass=abc.ABCMeta):
    """Interface for catalog of expressions that are interchangeable. Only one of
    them defines the specification. They are designed to be
    modified algorithmically.
    """

    def __init__(self, name):
        if SEPARATOR in name or SELECTION_SEPARATOR in name:
            error_msg = (
                f'Invalid name: {name}. Cannot contain characters '
                f'{SELECTION_SEPARATOR} or {SELECTION_SEPARATOR}'
            )
            raise excep.BiogemeError(error_msg)
        super().__init__()
        self.name = name

    @abc.abstractmethod
    def selected(self):
        """Return the selected expression and its name

        :return: the name and the selected expression
        :rtype: tuple(str, biogeme.expressions.Expression)
        """

    @abc.abstractmethod
    def get_iterator(self):
        """Returns an iterator on NamedExpression"""

    def catalog_size(self):
        """Provide the size of the catalog

        :return: number of expressions in the catalog
        :rtype: int
        """
        the_iterator = self.get_iterator()

        return len(list(the_iterator))

    def selected_name(self):
        """Obtain the name of the selection

        :return: the name of the selected expression
        :rtype: str
        """
        name, _ = self.selected()
        return name

    def selected_expression(self):
        """Obtain the selected expression

        :return: the selected expression
        :rtype: biogeme.expressions.Expression
        """
        _, the_expression = self.selected()
        return the_expression

    def getValue(self):
        """Evaluates the value of the expression

        :return: value of the expression
        :rtype: float
        """
        _, expr = self.selected()
        return expr.getValue()

    def get_id(self):
        """Retrieve the id of the expression used in the signature

        :return: id of the object
        :rtype: int
        """
        _, expr = self.selected()
        return expr.get_id()

    def getSignature(self):
        """The signature of a string characterizing an expression.

        This is designed to be communicated to C++, so that the
        expression can be reconstructed in this environment.

        The list contains the following elements:

            1. the signatures of all the children expressions,
            2. the name of the expression between < >
            3. the id of the expression between { }
            4. the number of children between ( )
            5. the ids of each children, preceeded by a comma.

        Consider the following expression:

        .. math:: 2 \\beta_1  V_1 -
            \\frac{\\exp(-\\beta_2 V_2) }
            { \\beta_3  (\\beta_2 \\geq \\beta_1)}.

        It is defined as::

            2 * beta1 * Variable1 - expressions.exp(-beta2*Variable2) /
                 (beta3 * (beta2 >= beta1))

        And its signature is::

            [b'<Numeric>{4780527008},2',
             b'<Beta>{4780277152}"beta1"[0],0,0',
             b'<Times>{4780526952}(2),4780527008,4780277152',
             b'<Variable>{4511837152}"Variable1",5,2',
             b'<Times>{4780527064}(2),4780526952,4511837152',
             b'<Beta>{4780277656}"beta2"[0],1,1',
             b'<UnaryMinus>{4780527120}(1),4780277656',
             b'<Variable>{4511837712}"Variable2",6,3',
             b'<Times>{4780527176}(2),4780527120,4511837712',
             b'<exp>{4780527232}(1),4780527176',
             b'<Beta>{4780277264}"beta3"[1],2,0',
             b'<Beta>{4780277656}"beta2"[0],1,1',
             b'<Beta>{4780277152}"beta1"[0],0,0',
             b'<GreaterOrEqual>{4780527288}(2),4780277656,4780277152',
             b'<Times>{4780527344}(2),4780277264,4780527288',
             b'<Divide>{4780527400}(2),4780527232,4780527344',
             b'<Minus>{4780527456}(2),4780527064,4780527400']

        :return: list of the signatures of an expression and its children.
        :rtype: list(string)

        """
        _, expr = self.selected()
        return expr.getSignature()

    def get_children(self):
        """Retrieve the list of children

        :return: list of children
        :rtype: list(Expression)
        """
        _, expr = self.selected()
        return expr.get_children()

    def __str__(self):
        named_expression = self.selected()
        return f'[{self.name}: {named_expression.name}]{named_expression.expression}'

    def setIdManager(self, id_manager):
        """The ID manager contains the IDs of the elementary expressions.

        It is externally created, as it may nee to coordinate the
        numbering of several expressions. It is stored only in the
        expressions of type Elementary.

        :param id_manager: ID manager to be propagated to the
            elementary expressions. If None, all the IDs are set to None.
        :type id_manager: class IdManager
        """
        _, expr = self.selected()
        expr.setIdManager(id_manager)

    def check_panel_trajectory(self):
        """Set of variables defined outside of 'PanelLikelihoodTrajectory'

        :return: List of names of variables
        :rtype: set(str)
        """
        _, expr = self.selected()
        return expr.check_panel_trajectory()

    def check_draws(self):
        """Set of draws defined outside of 'MonteCarlo'

        :return: List of names of variables
        :rtype: set(str)
        """
        _, expr = self.selected()
        return expr.check_draws()

    def check_rv(self):
        """Set of random variables defined outside of 'Integrate'

        :return: List of names of variables
        :rtype: set(str)
        """
        _, expr = self.selected()
        return expr.check_rv()

    def getStatusIdManager(self):
        """Check the elementary expressions that are associated with
        an ID manager.

        :return: two sets of elementary expressions, those with and
            without an ID manager.
        :rtype: tuple(set(str), set(str))
        """
        _, expr = self.selected()
        return expr.getStatusIdManager()

    def getElementaryExpression(self, name):
        """Return: an elementary expression from its name if it appears in the
        expression.

        :param name: name of the elementary expression.
        :type name: string

        :return: the expression if it exists. None otherwise.
        :rtype: biogeme.expressions.Expression
        """
        _, expr = self.selected()
        return expr.getElementaryExpression(name)

    def set_of_elementary_expression(self, the_type):
        """Extract a dict with all elementary expressions of a specific type

        :param the_type: the type of expression
        :type  the_type: TypeOfElementaryExpression

        :return: returns a set with the names of the elementary expressions
        :rtype: set(string.Expression)

        """
        _, expr = self.selected()
        return expr.set_of_elementary_expression(the_type)

    def dict_of_elementary_expression(self, the_type):
        """Extract a dict with all elementary expressions of a specific type

        :param the_type: the type of expression
        :type  the_type: TypeOfElementaryExpression

        :return: returns a dict with the variables appearing in the
               expression the keys being their names.
        :rtype: dict(string:biogeme.expressions.Expression)

        """
        _, expr = self.selected()
        return expr.dict_of_elementary_expression(the_type)

    def rename_elementary(self, names, prefix=None, suffix=None):
        """Rename elementary expressions by adding a prefix and/or a suffix

        :param names: names of expressions to rename
        :type names: list(str)

        :param prefix: if not None, the expression is renamed, with a
            prefix defined by this argument.
        :type prefix: str

        :param suffix: if not None, the expression is renamed, with a
            suffix defined by this argument.
        :type suffix: str
        """
        _, expr = self.selected()
        return expr.rename_elementary(names, prefix, suffix)

    def fix_betas(self, beta_values, prefix=None, suffix=None):
        """Fix all the values of the beta parameters appearing in the
        dictionary

        :param beta_values: dictionary containing the betas to be
            fixed (as key) and their value.
        :type beta_values: dict(str: float)

        :param prefix: if not None, the parameter is renamed, with a
            prefix defined by this argument.
        :type prefix: str

        :param suffix: if not None, the parameter is renamed, with a
            suffix defined by this argument.
        :type suffix: str

        """
        _, expr = self.selected()
        return expr.fix_betas(beta_values, prefix, suffix)

    def embedExpression(self, t):
        """Check if the expression contains an expression of type t.

        Typically, this would be used to check that a MonteCarlo
        expression contains a bioDraws expression.

        :return: True if the expression contains an expression of type t.
        :rtype: bool

        """
        _, expr = self.selected()
        return expr.embedExpression(t)

    def countPanelTrajectoryExpressions(self):
        """Count the number of times the PanelLikelihoodTrajectory
        is used in the formula. It should trigger an error if it
        is used more than once.

        :return: number of times the PanelLikelihoodTrajectory
            is used in the formula
        :rtype: int
        """
        _, expr = self.selected()
        return expr.countPanelTrajectoryExpressions()

    def change_init_values(self, betas):
        """Modifies the initial values of the Beta parameters.

        The fact that the parameters are fixed or free is irrelevant here.

        :param betas: dictionary where the keys are the names of the
                      parameters, and the values are the new value for
                      the parameters.
        :type betas: dict(string:float)
        """
        _, expr = self.selected()
        expr.change_init_values(betas)



import os
import unittest
import pandas as pd
import biogeme.database as db
import biogeme.biogeme as bio
import biogeme.distributions as dist
from biogeme import models
from biogeme.expressions import (
    Beta,
    log,
    RandomVariable,
    Integrate,
)

myPath = os.path.dirname(os.path.abspath(__file__))
df = pd.read_csv(f'{myPath}/swissmetro.dat', sep='\t')
database = db.Database('swissmetro', df)

# The Pandas data structure is available as database.data. Use all the
# Pandas functions to invesigate the database
# print(database.data.describe())

globals().update(database.variables)

# Here we use the 'biogeme' way for backward compatibility
exclude = ((PURPOSE != 1) * (PURPOSE != 3) + (CHOICE == 0)) > 0
database.remove(exclude)


ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_TRAIN = Beta('ASC_TRAIN', 0, None, None, 0)
ASC_SM = Beta('ASC_SM', 0, None, None, 1)
B_TIME = Beta('B_TIME', 0, None, None, 0)
B_TIME_S = Beta('B_TIME_S', 1, None, None, 0)
B_COST = Beta('B_COST', 0, None, None, 0)

# Define a random parameter, normally distirbuted, designed to be used
# for Monte-Carlo simulation

omega = RandomVariable('omega')
density = dist.normalpdf(omega)
B_TIME_RND = B_TIME + B_TIME_S * omega


# Utility functions

# If the person has a GA (season ticket) her incremental cost is actually 0
# rather than the cost value gathered from the
# network data.
SM_COST = SM_CO * (GA == 0)
TRAIN_COST = TRAIN_CO * (GA == 0)

# For numerical reasons, it is good practice to scale the data to
# that the values of the parameters are around 1.0.
# A previous estimation with the unscaled data has generated
# parameters around -0.01 for both cost and time. Therefore, time and
# cost are multipled my 0.01.

TRAIN_TT_SCALED = database.DefineVariable('TRAIN_TT_SCALED', TRAIN_TT / 100.0)
TRAIN_COST_SCALED = database.DefineVariable(
    'TRAIN_COST_SCALED', TRAIN_COST / 100
)
SM_TT_SCALED = database.DefineVariable('SM_TT_SCALED', SM_TT / 100.0)
SM_COST_SCALED = database.DefineVariable('SM_COST_SCALED', SM_COST / 100)
CAR_TT_SCALED = database.DefineVariable('CAR_TT_SCALED', CAR_TT / 100)
CAR_CO_SCALED = database.DefineVariable('CAR_CO_SCALED', CAR_CO / 100)

V1 = ASC_TRAIN + B_TIME_RND * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED
V2 = ASC_SM + B_TIME_RND * SM_TT_SCALED + B_COST * SM_COST_SCALED
V3 = ASC_CAR + B_TIME_RND * CAR_TT_SCALED + B_COST * CAR_CO_SCALED

# Associate utility functions with the numbering of alternatives
V = {1: V1, 2: V2, 3: V3}

# Associate the availability conditions with the alternatives

CAR_AV_SP = database.DefineVariable('CAR_AV_SP', CAR_AV * (SP != 0))
TRAIN_AV_SP = database.DefineVariable('TRAIN_AV_SP', TRAIN_AV * (SP != 0))

av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

# The choice model is a logit, with availability conditions
condprob = models.logit(V, av, CHOICE)
prob = Integrate(condprob * density, 'omega')
logprob = log(prob)


class test_02(unittest.TestCase):
    def testEstimation(self):
        biogeme = bio.BIOGEME(database, logprob)
        biogeme.saveIterations = False
        biogeme.generateHtml = False
        biogeme.generatePickle = False
        results = biogeme.estimate()
        self.assertAlmostEqual(results.data.logLike, -5214.879, 2)


if __name__ == '__main__':
    unittest.main()



import os
import shutil
import tempfile
import unittest
import pandas as pd
import biogeme.database as db
import biogeme.biogeme as bio
from biogeme import models
from biogeme.expressions import Beta, Variable

myPath = os.path.dirname(os.path.abspath(__file__))
df = pd.read_csv(f'{myPath}/swissmetro.dat', sep='\t')
database = db.Database('swissmetro', df)

# The Pandas data structure is available as database.data. Use all the
# Pandas functions to invesigate the database
# print(database.data.describe())

PURPOSE = Variable('PURPOSE')
CHOICE = Variable('CHOICE')
GA = Variable('GA')
TRAIN_CO = Variable('TRAIN_CO')
CAR_AV = Variable('CAR_AV')
SP = Variable('SP')
TRAIN_AV = Variable('TRAIN_AV')
TRAIN_TT = Variable('TRAIN_TT')
SM_TT = Variable('SM_TT')
CAR_TT = Variable('CAR_TT')
CAR_CO = Variable('CAR_CO')
SM_CO = Variable('SM_CO')
SM_AV = Variable('SM_AV')

# Here we use the 'biogeme' way for backward compatibility
exclude = ((PURPOSE != 1) * (PURPOSE != 3) + (CHOICE == 0)) > 0
database.remove(exclude)


ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_TRAIN = Beta('ASC_TRAIN', 0, None, None, 0)
ASC_SM = Beta('ASC_SM', 0, None, None, 1)
B_TIME = Beta('B_TIME', 0, None, None, 0)
B_COST = Beta('B_COST', 0, None, None, 0)

SM_COST = SM_CO * (GA == 0)
TRAIN_COST = TRAIN_CO * (GA == 0)

CAR_AV_SP = database.DefineVariable('CAR_AV_SP', CAR_AV * (SP != 0))
TRAIN_AV_SP = database.DefineVariable('TRAIN_AV_SP', TRAIN_AV * (SP != 0))

TRAIN_TT_SCALED = database.DefineVariable('TRAIN_TT_SCALED', TRAIN_TT / 100.0)
TRAIN_COST_SCALED = database.DefineVariable(
    'TRAIN_COST_SCALED', TRAIN_COST / 100
)
SM_TT_SCALED = database.DefineVariable('SM_TT_SCALED', SM_TT / 100.0)
SM_COST_SCALED = database.DefineVariable('SM_COST_SCALED', SM_COST / 100)
CAR_TT_SCALED = database.DefineVariable('CAR_TT_SCALED', CAR_TT / 100)
CAR_CO_SCALED = database.DefineVariable('CAR_CO_SCALED', CAR_CO / 100)

V1 = ASC_TRAIN + B_TIME * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED
V2 = ASC_SM + B_TIME * SM_TT_SCALED + B_COST * SM_COST_SCALED
V3 = ASC_CAR + B_TIME * CAR_TT_SCALED + B_COST * CAR_CO_SCALED

# Associate utility functions with the numbering of alternatives
V = {1: V1, 2: V2, 3: V3}


# Associate the availability conditions with the alternatives

av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}


class test_01(unittest.TestCase):
    def setUp(self):
        """Create the configuration files"""
        # Create a temporary directory
        self.test_dir = tempfile.mkdtemp()
        self.scipy_file = os.path.join(self.test_dir, 'scipy.toml')
        with open(self.scipy_file, 'w', encoding='utf-8') as f:
            print('[Estimation]', file=f)
            print('optimization_algorithm = "scipy"', file=f)
            print('[MonteCarlo]', file=f)
            print('seed = 10', file=f)
        self.ls_file = os.path.join(self.test_dir, 'line_search.toml')
        with open(self.ls_file, 'w', encoding='utf-8') as f:
            print('[Estimation]', file=f)
            print('optimization_algorithm = "LS-newton"', file=f)
            print('[MonteCarlo]', file=f)
            print('seed = 10', file=f)
        self.tr_file = os.path.join(self.test_dir, 'trust_region.toml')
        with open(self.tr_file, 'w', encoding='utf-8') as f:
            print('[Estimation]', file=f)
            print('optimization_algorithm = "TR-newton"', file=f)
            print('[MonteCarlo]', file=f)
            print('seed = 10', file=f)
        self.simple_bounds_file = os.path.join(
            self.test_dir, 'simple_bounds.toml'
        )
        with open(self.simple_bounds_file, 'w', encoding='utf-8') as f:
            print('[Estimation]', file=f)
            print('optimization_algorithm = "simple_bounds"', file=f)
            print('[MonteCarlo]', file=f)
            print('seed = 10', file=f)

    def testEstimationScipy(self):
        logprob = models.loglogit(V, av, CHOICE)
        biogeme = bio.BIOGEME(
            database, logprob, parameter_file=self.scipy_file
        )
        biogeme.modelName = 'test_01'
        biogeme.generate_html = False
        biogeme.generate_pickle = False
        biogeme.saveIterations = False
        biogeme.bootstrap_samples = 10
        results = biogeme.estimate(run_bootstrap=True)
        self.assertAlmostEqual(results.data.logLike, -5331.252, 2)

    def testEstimationLineSearch(self):
        logprob = models.loglogit(V, av, CHOICE)
        biogeme = bio.BIOGEME(database, logprob, parameter_file=self.ls_file)
        biogeme.modelName = 'test_01'
        biogeme.generate_html = False
        biogeme.generate_pickle = False
        biogeme.saveIterations = False
        biogeme.bootstrap_samples=10
        results = biogeme.estimate(run_bootstrap=True)
        self.assertAlmostEqual(results.data.logLike, -5331.252, 2)

    def testEstimationTrustRegion(self):
        logprob = models.loglogit(V, av, CHOICE)
        biogeme = bio.BIOGEME(database, logprob, parameter_file=self.tr_file)
        biogeme.modelName = 'test_01'
        biogeme.generate_html = False
        biogeme.generate_pickle = False
        biogeme.saveIterations = False
        biogeme.bootstrap_samples=10
        results = biogeme.estimate(run_bootstrap=True)
        self.assertAlmostEqual(results.data.logLike, -5331.252, 2)

    def tearDown(self):
        # Remove the directory after the test
        shutil.rmtree(self.test_dir)


if __name__ == '__main__':
    unittest.main()



"""Scripts deleting Biogeme output files

:author: Michel Bierlaire
:date: Sun Jul 30 15:58:57 2023
"""
import os
import sys

EXTENSIONS_TO_CLEAN = ['html', 'pickle', 'iter', 'log', 'pareto', '.tex', '.F12', '~', '#']


def file_to_erase(the_filename: str) -> bool:
    """Checks if a file must be erased

    :param the_filename: name fo the file
    :type the_filename: str

    :return: True if the file must be erased. False otherwise.
    :rtype: bool
    """
    for ext in EXTENSIONS_TO_CLEAN:
        if the_filename.endswith(ext):
            return True

    return False


def get_file_list():
    """Retrieves the list of files to be erased.

    :return: list of file names
    :rtype: list[str]
    """
    
    the_list = []
    current_directory = os.getcwd()
    for dirpath, _, filenames in os.walk(current_directory):
        for filename in filenames:
            if file_to_erase(filename):
                file_path = os.path.join(dirpath, filename)
                the_list.append(file_path)
    return the_list


def ask_for_confirmation(the_list):
    """Function asking confirmation to the user.

    :param the_list: luist of files to erase
    :type the_list: list[str]

    :return: True if OK to proceed, False otherwise.
    :rtype: bool
    """
    while True:
        print('This script will erase the following files:')
        print('\n'.join(the_list))
        text = 'Do you want to proceed? (yes/no): '
        response = input(text).strip().lower()
        if response in ('yes', 'y'):
            return True
        if response in ('no', 'n'):
            return False
        print("Invalid response. Please enter 'yes' or 'no'.")


file_list = get_file_list()
if not file_list:
    print('No file to be cleaned.')
    sys.exit(0)


if ask_for_confirmation(file_list):
    for file_name in file_list:
        os.remove(file_name)
    if len(file_list) == 1:
        print('One file has been erased.')
    else:
        print(f'{len(file_list)} files have been erased.')

else:
    print("Script execution aborted.")



"""File b07discrete_mixture.py

:author: Michel Bierlaire, EPFL
:date: Sun Apr  9 17:57:07 2023

 Example of a discrete mixture of logit (or latent class model)

"""

import biogeme.biogeme as bio
from biogeme import models
from biogeme.expressions import Beta, log

from swissmetro_data import (
    database,
    CHOICE,
    SM_AV,
    CAR_AV_SP,
    TRAIN_AV_SP,
    TRAIN_TT_SCALED,
    TRAIN_COST_SCALED,
    SM_TT_SCALED,
    SM_COST_SCALED,
    CAR_TT_SCALED,
    CAR_CO_SCALED,
)


# Parameters to be estimated
ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_TRAIN = Beta('ASC_TRAIN', 0, None, None, 0)
ASC_SM = Beta('ASC_SM', 0, None, None, 1)
B_TIME = Beta('B_TIME', 0, None, None, 0)
B_COST = Beta('B_COST', 0, None, None, 0)

PROB_CLASS1 = Beta('PROB_CLASS1', 0.5, 0, 1, 0)
PROB_CLASS2 = 1 - PROB_CLASS1

# Definition of the utility functions for latent class 1, where the
# time coefficient is zero
V11 = ASC_TRAIN + B_COST * TRAIN_COST_SCALED
V12 = ASC_SM + B_COST * SM_COST_SCALED
V13 = ASC_CAR + B_COST * CAR_CO_SCALED

# Associate utility functions with the numbering of alternatives
V1 = {1: V11, 2: V12, 3: V13}

# Definition of the utility functions for latent class 2, whete the
# time coefficient is estimated
V21 = ASC_TRAIN + B_TIME * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED
V22 = ASC_SM + B_TIME * SM_TT_SCALED + B_COST * SM_COST_SCALED
V23 = ASC_CAR + B_TIME * CAR_TT_SCALED + B_COST * CAR_CO_SCALED

# Associate utility functions with the numbering of alternatives
V2 = {1: V21, 2: V22, 3: V23}

# Associate the availability conditions with the alternatives
av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}


# The choice model is a discrete mixture of logit, with availability conditions
prob1 = models.logit(V1, av, CHOICE)
prob2 = models.logit(V2, av, CHOICE)
prob = PROB_CLASS1 * prob1 + PROB_CLASS2 * prob2
logprob = log(prob)

# Create the Biogeme object
the_biogeme = bio.BIOGEME(database, logprob)
the_biogeme.modelName = 'b07discrete_mixture'

# Estimate the parameters
results = the_biogeme.estimate()
print(results.short_summary())
pandas_results = results.getEstimatedParameters()
print(pandas_results)



"""File scenarios.py

:author: Michel Bierlaire, EPFL
:date: Wed Apr 12 20:51:37 2023

 Specification of a nested logit model, that will be estimated, and
 used for simulation.  Three alternatives: public transporation, car
 and slow modes.  RP data.
 Based on the Optima data.
 It contains a function that generates scenarios where the current
 cost of public transportation is multiplied by a factor.

"""

from biogeme.expressions import Beta
from optima_data import (
    Choice,
    TimePT,
    TimeCar,
    CostCarCHF,
    MarginalCostPT,
    distance_km,
    Gender,
    OccupStat,
)

# List of parameters to be estimated
ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_PT = Beta('ASC_PT', 0, None, None, 1)
ASC_SM = Beta('ASC_SM', 0, None, None, 0)
BETA_TIME_FULLTIME = Beta('BETA_TIME_FULLTIME', 0, None, None, 0)
BETA_TIME_OTHER = Beta('BETA_TIME_OTHER', 0, None, None, 0)
BETA_DIST_MALE = Beta('BETA_DIST_MALE', 0, None, None, 0)
BETA_DIST_FEMALE = Beta('BETA_DIST_FEMALE', 0, None, None, 0)
BETA_DIST_UNREPORTED = Beta('BETA_DIST_UNREPORTED', 0, None, None, 0)
BETA_COST = Beta('BETA_COST', 0, None, None, 0)

# Definition of variables:
# For numerical reasons, it is good practice to scale the data to
# that the values of the parameters are around 1.0.

TimePT_scaled = TimePT / 200
TimeCar_scaled = TimeCar / 200
CostCarCHF_scaled = CostCarCHF / 10
distance_km_scaled = distance_km / 5
male = Gender == 1
female = Gender == 2
unreportedGender = Gender == -1

fulltime = OccupStat == 1
notfulltime = OccupStat != 1


def scenario(factor=1.0):
    """Provide the model specification for a scenario with the price of
        public transportation is multiplied by a factor

    :param factor: factor that multiples the price of public transportation.
    :type factor: float

    :return: a dict with the utility functions, the nesting structure,
        and the choice expression.

    :rtype: dict(int: biogeme.expression), tuple(biogeme.expression,
        list(int)), biogeme.expression
    """
    marginal_cost_scenario = MarginalCostPT * factor
    marginal_cost_pt_scaled = marginal_cost_scenario / 10
    # Definition of utility functions:
    v_pt = (
        ASC_PT
        + BETA_TIME_FULLTIME * TimePT_scaled * fulltime
        + BETA_TIME_OTHER * TimePT_scaled * notfulltime
        + BETA_COST * marginal_cost_pt_scaled
    )
    v_car = (
        ASC_CAR
        + BETA_TIME_FULLTIME * TimeCar_scaled * fulltime
        + BETA_TIME_OTHER * TimeCar_scaled * notfulltime
        + BETA_COST * CostCarCHF_scaled
    )
    v_sm = (
        ASC_SM
        + BETA_DIST_MALE * distance_km_scaled * male
        + BETA_DIST_FEMALE * distance_km_scaled * female
        + BETA_DIST_UNREPORTED * distance_km_scaled * unreportedGender
    )

    # Associate utility functions with the numbering of alternatives
    V = {0: v_pt, 1: v_car, 2: v_sm}

    # Definition of the nests:
    # 1: nests parameter
    # 2: list of alternatives
    mu_nocar = Beta('mu_nocar', 1.0, 1.0, None, 0)

    car_nest = 1.0, [1]
    no_car_nest = mu_nocar, [0, 2]
    nests = car_nest, no_car_nest
    return V, nests, Choice, marginal_cost_scenario



"""
Simple estimation test from the Ben-Akiva and Lerman book

:author: Michel Bierlaire
:date: Wed Apr 29 18:43:34 2020

"""
import unittest
from os import path
import shutil
import tempfile

import pandas as pd

import biogeme.database as db
import biogeme.biogeme as bio
from biogeme import models
from biogeme.expressions import Variable, Beta

GROUP = Variable('GROUP')
SURVEY = Variable('SURVEY')
SP = Variable('SP')
ID = Variable('ID')
PURPOSE = Variable('PURPOSE')
FIRST = Variable('FIRST')
TICKET = Variable('TICKET')
WHO = Variable('WHO')
LUGGAGE = Variable('LUGGAGE')
AGE = Variable('AGE')
MALE = Variable('MALE')
INCOME = Variable('INCOME')
GA = Variable('GA')
ORIGIN = Variable('ORIGIN')
DEST = Variable('DEST')
TRAIN_AV = Variable('TRAIN_AV')
CAR_AV = Variable('CAR_AV')
SM_AV = Variable('SM_AV')
TRAIN_TT = Variable('TRAIN_TT')
TRAIN_CO = Variable('TRAIN_CO')
TRAIN_HE = Variable('TRAIN_HE')
SM_TT = Variable('SM_TT')
SM_CO = Variable('SM_CO')
SM_HE = Variable('SM_HE')
SM_SEATS = Variable('SM_SEATS')
CAR_TT = Variable('CAR_TT')
CAR_CO = Variable('CAR_CO')
CHOICE = Variable('CHOICE')


class test_biogeme(unittest.TestCase):
    def setUp(self):
        # Create a temporary directory
        self.test_dir = tempfile.mkdtemp()
        self.toml_file = path.join(self.test_dir, 'biogeme.toml')

        data = {
            'ID': pd.Series([i + 1 for i in range(21)]),
            'AutoTime': pd.Series(
                [
                    52.9,
                    4.1,
                    4.1,
                    56.2,
                    51.8,
                    0.2,
                    27.6,
                    89.9,
                    41.5,
                    95.0,
                    99.1,
                    18.5,
                    82.0,
                    8.6,
                    22.5,
                    51.4,
                    81.0,
                    51.0,
                    62.2,
                    95.1,
                    41.6,
                ]
            ),
            'TransitTime': pd.Series(
                [
                    4.4,
                    28.5,
                    86.9,
                    31.6,
                    20.2,
                    91.2,
                    79.7,
                    2.2,
                    24.5,
                    43.5,
                    8.4,
                    84.0,
                    38.0,
                    1.6,
                    74.1,
                    83.8,
                    19.2,
                    85.0,
                    90.1,
                    22.2,
                    91.5,
                ]
            ),
            'Choice': pd.Series(
                [1, 1, 0, 1, 1, 0, 0, 1, 1, 1, 1, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0]
            ),
        }
        pandas = pd.DataFrame(data)

        self.database = db.Database('akiva', pandas)

    def tearDown(self):
        # Remove the directory after the test
        shutil.rmtree(self.test_dir)

    def testEstimation(self):
        AutoTime = Variable('AutoTime')
        TransitTime = Variable('TransitTime')
        Choice = Variable('Choice')

        ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
        B_TIME = Beta('B_TIME', 0, None, None, 0)

        V = {0: ASC_CAR + B_TIME * AutoTime, 1: B_TIME * TransitTime}

        av = {0: 1, 1: 1}

        logprob = models.loglogit(V, av, Choice)

        biogeme = bio.BIOGEME(self.database, logprob, parameter_file=self.toml_file)
        biogeme.modelName = 'test'
        biogeme.generate_html = False
        biogeme.generate_pickle = False
        biogeme.saveIterations = False
        results = biogeme.estimate()
        self.assertAlmostEqual(results.data.logLike, -6.166042, 2)


if __name__ == '__main__':
    unittest.main()



"""File b24halton_mixture.py

:author: Michel Bierlaire, EPFL
:date: Wed Apr 12 18:21:13 2023

 Example of a mixture of logit models, using quasi Monte-Carlo integration with
 Halton draws (base 5).
 The mixing distribution is normal.

"""

import biogeme.logging as blog
import biogeme.biogeme as bio
from biogeme import models

from biogeme.expressions import Beta, bioDraws, MonteCarlo, log

from swissmetro_data import (
    database,
    CHOICE,
    CAR_AV_SP,
    TRAIN_AV_SP,
    TRAIN_TT_SCALED,
    TRAIN_COST_SCALED,
    SM_TT_SCALED,
    SM_COST_SCALED,
    CAR_TT_SCALED,
    CAR_CO_SCALED,
    SM_AV,
)

logger = blog.get_screen_logger(level=blog.INFO)
logger.info('Example b24halton_mixture.py')


# Parameters to be estimated
ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_TRAIN = Beta('ASC_TRAIN', 0, None, None, 0)
ASC_SM = Beta('ASC_SM', 0, None, None, 1)
B_COST = Beta('B_COST', 0, None, None, 0)

# Define a random parameter, normally distributed, designed to be used
# for Monte-Carlo simulation
B_TIME = Beta('B_TIME', 0, None, None, 0)

# It is advised not to use 0 as starting value for the following parameter.
B_TIME_S = Beta('B_TIME_S', 1, None, None, 0)
# Define a random parameter with a normal distribution, designed to be used
# for quasi Monte-Carlo simulation with Halton draws (base 5).
B_TIME_RND = B_TIME + B_TIME_S * bioDraws('B_TIME_RND', 'NORMAL_HALTON5')

# Definition of the utility functions
V1 = ASC_TRAIN + B_TIME_RND * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED
V2 = ASC_SM + B_TIME_RND * SM_TT_SCALED + B_COST * SM_COST_SCALED
V3 = ASC_CAR + B_TIME_RND * CAR_TT_SCALED + B_COST * CAR_CO_SCALED

# Associate utility functions with the numbering of alternatives
V = {1: V1, 2: V2, 3: V3}

# Associate the availability conditions with the alternatives
av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

# Conditional to B_TIME_RND, we have a logit model (called the kernel)
prob = models.logit(V, av, CHOICE)

# We integrate over B_TIME_RND using Monte-Carlo
logprob = log(MonteCarlo(prob))

# These notes will be included as such in the report file.
USER_NOTES = (
    'Example of a mixture of logit models with three alternatives, '
    'approximated using Monte-Carlo integration with Halton draws.'
)

# Create the Biogeme object
the_biogeme = bio.BIOGEME(database, logprob, userNotes=USER_NOTES)
the_biogeme.modelName = 'b24halton_mixture'

# Estimate the parameters
results = the_biogeme.estimate()
print(results.short_summary())
pandas_results = results.getEstimatedParameters()
print(pandas_results)



"""Represents the configuration of a multiple expression

:author: Michel Bierlaire
:date: Sun Apr  2 14:15:17 2023

"""

import logging
from typing import NamedTuple
import biogeme.exceptions as excep

logger = logging.getLogger(__name__)


class SelectionTuple(NamedTuple):
    controller: str
    selection: str


SEPARATOR = ';'
SELECTION_SEPARATOR = ':'


class Configuration:
    """Represents the configuration of a multiple expression. It is
    internally represented as a sorted list of tuples.

    """

    def __init__(self, selections=None):
        """Ctor.

        :param selections: list of tuples, where each of
            them associates a controller name with the selected configuration
        :type list_of_selections: list(SelectionTuple)

        """
        if selections is None:
            self.__selections = None
        else:
            self.selections = selections

    @property
    def selections(self):
        return self.__selections

    @selections.setter
    def selections(self, value):
        self.__selections = sorted(value)
        self.__check_list_validity()
        self.string_id = self.get_string_id()

    @classmethod
    def from_string(cls, string_id):
        """Ctor from a string representation

        :param string_id: string ID
        :type string_id: str

        """
        terms = string_id.split(SEPARATOR)
        the_config = {}
        for term in terms:
            try:
                controller, selection = term.split(SELECTION_SEPARATOR)
            except ValueError as exc:
                error_msg = (
                    f'{exc}: Invalid syntax for ID {term}. Expecting a separator '
                    f'[{SELECTION_SEPARATOR}]'
                )
                raise excep.BiogemeError(error_msg)

            the_config[controller] = selection
        return cls.from_dict(the_config)

    @classmethod
    def from_dict(cls, dict_of_selections):
        """Ctor from dict

        :param dict_of_selections: dict associating a catalog name
            with the selected configuration
        :type dict_of_selections: dict(str: str)

        """
        the_list = (
            SelectionTuple(controller=controller, selection=selection)
            for controller, selection in dict_of_selections.items()
        )
        return cls(selections=the_list)

    @classmethod
    def from_tuple_of_configurations(cls, tuple_of_configurations):
        """Ctor from tuple of configurations that are merged together.

        In the presence of two different selections for the same
        catalog, the first one is selected,
        and the others ignored.

        :param tuple_of_configurations: tuple of configurations to merge
        :type tuple_of_configurations: tuple(Configuration)

        """
        known_controllers = set()
        selections = []
        for configuration in tuple_of_configurations:
            if configuration.selections is not None:
                for selection in configuration.selections:
                    if selection.controller not in known_controllers:
                        selections.append(selection)
                        known_controllers.add(selection.controller)
        return cls(selections)

    def set_of_controllers(self):
        return {selection.controller for selection in self.selections}

    def __eq__(self, other):
        return self.string_id == other.string_id

    def __hash__(self):
        return hash(self.string_id)

    def __repr__(self):
        return repr(self.string_id)

    def __str__(self):
        return str(self.string_id)

    def get_string_id(self):
        """The string ID is a unique string representation of the configuration

        :return: string ID
        :rtype: str
        """
        terms = [
            f'{selection.controller}{SELECTION_SEPARATOR}{selection.selection}'
            for selection in self.selections
        ]
        return SEPARATOR.join(terms)

    def get_html(self):
        html = '<p>Specification</p><p><ul>\n'
        for selection_tuple in self.selections:
            html += (
                f'<li>{selection_tuple.controller}: '
                f'{selection_tuple.selection}</li>\n'
            )
        html += '</ul></p>\n'
        return html

    def get_selection(self, controller_name):
        """Retrieve the selection of a given controller

        :param controller_name: name of the controller
        :type controller_name: str

        :return: name of the selected config, or None if controller is not known
        :rtype: str
        """
        for selection in self.selections:
            if selection.controller == controller_name:
                return selection.selection
        return None

    def __check_list_validity(self):
        """Check the validity of the list.

        :raise BiogemeError: if the same catalog appears more than once in the list
        """
        unique_items = set()
        for item in self.__selections:
            if item.controller in unique_items:
                error_msg = (
                    f'Controller {item.controller} appears more than once in the '
                    f'configuration: {self.__selections}'
                )
                raise excep.BiogemeError(error_msg)
            unique_items.add(item.controller)



import os
import nbformat
from nbconvert.preprocessors import ExecutePreprocessor
from nbconvert import HTMLExporter

def execute_notebook(notebook_path):
    # Set environment variables to suppress the debugger warnings
    os.environ['PYDEVD_DISABLE_FILE_VALIDATION'] = '1'
    os.environ['PYDEVD_WARN_FROZEN'] = 'off'

    with open(notebook_path) as f:
        nb = nbformat.read(f, as_version=4)

    ep = ExecutePreprocessor(timeout=600)  # Adjust the timeout as needed (in seconds)

    try:
        ep.preprocess(nb, {'metadata': {'path': os.path.dirname(notebook_path)}})
    except Exception as e:
        raise RuntimeError(f"Error executing notebook: {notebook_path}, error: {str(e)}")
    finally:
        # Clean up the environment variables after execution
        os.environ.pop('PYDEVD_DISABLE_FILE_VALIDATION', None)
        os.environ.pop('PYDEVD_WARN_FROZEN', None)

    # Save the executed notebook
    with open(notebook_path, 'w', encoding='utf-8') as f:
        nbformat.write(nb, f)
        
    return nb

def export_html(notebook_path, output_dir):
    exporter = HTMLExporter()
    output_notebook, _ = exporter.from_filename(notebook_path)
    output_file = os.path.join(output_dir, os.path.splitext(os.path.basename(notebook_path))[0] + '.html')
    
    with open(output_file, 'w', encoding='utf-8') as f:
        f.write(output_notebook)

if __name__ == '__main__':
    notebooks_directory = '.'  # Replace with the directory containing your notebooks

    # Optional: Create a directory to store the executed notebook outputs in HTML format
    output_directory = 'output_directory'
    if not os.path.exists(output_directory):
        os.makedirs(output_directory)

    # List all .ipynb files in the notebooks directory and execute each of them
    for filename in os.listdir(notebooks_directory):
        if filename.endswith('.ipynb'):
            notebook_path = os.path.join(notebooks_directory, filename)
            try:
                execute_notebook(notebook_path)
                print(f"Notebook executed successfully: {notebook_path}")
                export_html(notebook_path, output_directory)  # Export notebook to HTML if needed
            except Exception as e:
                print(f"Error executing notebook: {notebook_path}, error: {str(e)}")
                continue



"""File b05normal_mixture_all_algos.py

:author: Michel Bierlaire, EPFL
:date: Sun Apr  9 17:38:34 2023

 Example of a mixture of logit models, using Monte-Carlo integration.
Estimation using several algorithms
"""

import biogeme.logging as blog
import biogeme.biogeme as bio
from biogeme import models
import biogeme.exceptions as excep
from biogeme.expressions import Beta, bioDraws, log, MonteCarlo
from swissmetro_data import (
    database,
    CHOICE,
    SM_AV,
    CAR_AV_SP,
    TRAIN_AV_SP,
    TRAIN_TT_SCALED,
    TRAIN_COST_SCALED,
    SM_TT_SCALED,
    SM_COST_SCALED,
    CAR_TT_SCALED,
    CAR_CO_SCALED,
)

logger = blog.get_screen_logger(level=blog.INFO)
logger.info('Example b05normal_mixture_all_algos.py')

# Parameters to be estimated
ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_TRAIN = Beta('ASC_TRAIN', 0, None, None, 0)
ASC_SM = Beta('ASC_SM', 0, None, None, 1)
B_COST = Beta('B_COST', 0, None, None, 0)

# Define a random parameter, normally distributed, designed to be used
# for Monte-Carlo simulation
B_TIME = Beta('B_TIME', 0, None, None, 0)

# It is advised not to use 0 as starting value for the following parameter.
B_TIME_S = Beta('B_TIME_S', 1, None, None, 0)
B_TIME_RND = B_TIME + B_TIME_S * bioDraws('B_TIME_RND', 'NORMAL')

# Definition of the utility functions
V1 = ASC_TRAIN + B_TIME_RND * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED
V2 = ASC_SM + B_TIME_RND * SM_TT_SCALED + B_COST * SM_COST_SCALED
V3 = ASC_CAR + B_TIME_RND * CAR_TT_SCALED + B_COST * CAR_CO_SCALED

# Associate utility functions with the numbering of alternatives
V = {1: V1, 2: V2, 3: V3}

# Associate the availability conditions with the alternatives
av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

# Conditional to B_TIME_RND, we have a logit model (called the kernel)
prob = models.logit(V, av, CHOICE)

# We integrate over B_TIME_RND using Monte-Carlo
logprob = log(MonteCarlo(prob))

algos = {
    'scipy                ': 'scipy',
    'Line search          ': 'LS-newton',
    'Trust region (dogleg)': 'TR-newton',
    'Trust region (cg)    ': 'TR-newton',
    'LS-BFGS              ': 'LS-BFGS',
    'TR-BFGS              ': 'TR-BFGS',
    'Simple bounds Newton fCG': 'simple_bounds',
    'Simple bounds BFGS fCG  ': 'simple_bounds',
    'Simple bounds hybrid fCG': 'simple_bounds',
    'Simple bounds Newton iCG': 'simple_bounds',
    'Simple bounds BFGS iCG  ': 'simple_bounds',
    'Simple bounds hybrid iCG': 'simple_bounds',
}

algoParameters = {
    'Trust region (dogleg)': {'dogleg': True},
    'Trust region (cg)    ': {'dogleg': False},
    'Simple bounds Newton fCG': {
        'second_derivatives': 1.0,
        'infeasible_cg': False,
    },
    'Simple bounds BFGS fCG  ': {
        'second_derivatives': 0.0,
        'infeasible_cg': False,
    },
    'Simple bounds hybrid fCG': {
        'second_derivatives': 0.5,
        'infeasible_cg': False,
    },
    'Simple bounds Newton iCG': {
        'second_derivatives': 1.0,
        'infeasible_cg': True,
    },
    'Simple bounds BFGS iCG  ': {
        'second_derivatives': 0.0,
        'infeasible_cg': True,
    },
    'Simple bounds hybrid iCG': {
        'second_derivatives': 0.5,
        'infeasible_cg': True,
    },
}

results = {}
msg = str()
for name, algo in algos.items():
    # Create the Biogeme object
    the_biogeme = bio.BIOGEME(database, logprob)
    the_biogeme.algorithm_name = algo
    the_biogeme.modelName = f'b05normal_mixture_all_algos_{name}'.strip()
    p = algoParameters.get(name)
    if p is not None:
        for attr, value in p.items():
            setattr(the_biogeme, attr, value)
    try:
        results[name] = the_biogeme.estimate()
        msg += (
            f'{name}\t{results[name].data.logLike:.2f}\t'
            f'{results[name].data.gradientNorm:.2g}\t'
            f'{results[name].data.optimizationMessages["Optimization time"]}'
            f'\t{results[name].data.optimizationMessages["Cause of termination"]}'
            f'\n'
        )
    except excep.BiogemeError as e:
        print(e)
        results[name] = None
        msg += f'{name}\tFailed to estimate the model'


SUMMARY_FILE = '05normalMixture_allAlgos.log'
with open(SUMMARY_FILE, 'w', encoding='utf-8') as f:
    print('Algorithm\t\tloglike\t\tnormg\ttime\t\tdiagnostic', file=f)
    print('+++++++++\t\t+++++++\t\t+++++\t++++\t\t++++++++++', file=f)
    print(msg, file=f)
print(f'Summary reported in file {SUMMARY_FILE}')



"""File optima.py

Data processing for the optima case study.

:author: Michel Bierlaire, EPFL
:date: Thu Apr 13 16:40:59 2023

"""


import pandas as pd
import biogeme.database as db
from biogeme import models
from biogeme.expressions import Variable


# Read the data
df = pd.read_csv('optima.dat', sep='\t')
database = db.Database('optima', df)

Choice = Variable('Choice')
TimePT = Variable('TimePT')
MarginalCostPT = Variable('MarginalCostPT')
TimeCar = Variable('TimeCar')
CostCarCHF = Variable('CostCarCHF')
distance_km = Variable('distance_km')
TripPurpose = Variable('TripPurpose')
WaitingTimePT = Variable('WaitingTimePT')
CalculatedIncome = Variable('CalculatedIncome')
age = Variable('age')
NbCar = Variable('NbCar')
NbBicy = Variable('NbBicy')
HouseType = Variable('HouseType')
Gender = Variable('Gender')
FamilSitu = Variable('FamilSitu')
GenAbST = Variable('GenAbST')
Education = Variable('Education')
Envir01 = Variable('Envir01')
Envir02 = Variable('Envir02')
Envir03 = Variable('Envir03')
Mobil11 = Variable('Mobil11')
Mobil14 = Variable('Mobil14')
Mobil16 = Variable('Mobil16')
Mobil17 = Variable('Mobil17')


# Exclude observations such that the chosen alternative is -1
database.remove(Choice == -1.0)

# Piecewise linear definition of income
ScaledIncome = database.DefineVariable('ScaledIncome', CalculatedIncome / 1000)

thresholds = [None, 4, 6, 8, 10, None]
formulaIncome = models.piecewiseFormula(
    ScaledIncome, thresholds, [0.0, 0.0, 0.0, 0.0, 0.0]
)

# Definition of other variables
age_65_more = database.DefineVariable('age_65_more', age >= 65)
moreThanOneCar = database.DefineVariable('moreThanOneCar', NbCar > 1)
moreThanOneBike = database.DefineVariable('moreThanOneBike', NbBicy > 1)
individualHouse = database.DefineVariable('individualHouse', HouseType == 1)
male = database.DefineVariable('male', Gender == 1)
haveChildren = database.DefineVariable(
    'haveChildren', ((FamilSitu == 3) + (FamilSitu == 4)) > 0
)
haveGA = database.DefineVariable('haveGA', GenAbST == 1)
highEducation = database.DefineVariable('highEducation', Education >= 6)



"""File airline_data.py

:author: Michel Bierlaire, EPFL
:date: Fri Mar 31 10:42:52 2023

Read data for the Airline example
"""

# Too constraining
# pylint: disable=invalid-name, undefined-variable

import pandas as pd
import biogeme.database as db
from biogeme.expressions import Variable

# Read the data
df = pd.read_csv('airline.dat', sep='\t')

# Update some data
df.loc[df['q17_Gender'] == 99, 'q17_Gender'] = -1
df.loc[df['q20_Education'] == 99, 'q20_Education'] = -1

# High education
df.loc[df['q20_Education'] <= 10, 'education'] = 1
# Medium education
df.loc[df['q20_Education'] <= 5, 'education'] = 2
# Low education
df.loc[df['q20_Education'] <= 2, 'education'] = 3
df.loc[df['q20_Education'] == 99, 'education'] = -1
df.loc[df['q20_Education'] == -1, 'education'] = -1

df.loc[
    df['q11_DepartureOrArrivalIsImportant'] == -1, 'q11_DepartureOrArrivalIsImportant'
] = 0

database = db.Database('airline', df)

ArrivalTimeHours_1 = Variable('ArrivalTimeHours_1')
BestAlternative_1 = Variable('BestAlternative_1')
BestAlternative_2 = Variable('BestAlternative_2')
BestAlternative_3 = Variable('BestAlternative_3')
q11_DepartureOrArrivalIsImportant = Variable('q11_DepartureOrArrivalIsImportant')
q11_DepartureOrArrivalIsImportant = Variable('q11_DepartureOrArrivalIsImportant')
q12_IdealDepTime = Variable('q12_IdealDepTime')
q13_IdealArrTime = Variable('q13_IdealArrTime')
DepartureTimeMins_1 = Variable('DepartureTimeMins_1')
ArrivalTimeMins_1 = Variable('ArrivalTimeMins_1')
DepartureTimeMins_2 = Variable('DepartureTimeMins_2')
ArrivalTimeMins_2 = Variable('ArrivalTimeMins_2')
DepartureTimeMins_3 = Variable('DepartureTimeMins_3')
ArrivalTimeMins_3 = Variable('ArrivalTimeMins_3')
Fare_1 = Variable('Fare_1')
Fare_2 = Variable('Fare_2')
Fare_3 = Variable('Fare_3')
Legroom_1 = Variable('Legroom_1')
Legroom_2 = Variable('Legroom_2')
Legroom_3 = Variable('Legroom_3')
TripTimeHours_1 = Variable('TripTimeHours_1')
TripTimeHours_2 = Variable('TripTimeHours_2')
TripTimeHours_3 = Variable('TripTimeHours_3')
q02_TripPurpose = Variable('q02_TripPurpose')
q03_WhoPays = Variable('q03_WhoPays')
q17_Gender = Variable('q17_Gender')
q20_Education = Variable('q20_Education')
education = Variable('education')


exclude = ArrivalTimeHours_1 == -1
database.remove(exclude)

chosenAlternative = (
    (BestAlternative_1 * 1) + (BestAlternative_2 * 2) + (BestAlternative_3 * 3)
)

DepartureTimeSensitive = database.DefineVariable(
    'DepartureTimeSensitive', q11_DepartureOrArrivalIsImportant == 1
)
ArrivalTimeSensitive = database.DefineVariable(
    'ArrivalTimeSensitive', q11_DepartureOrArrivalIsImportant == 2
)

DesiredDepartureTime = database.DefineVariable('DesiredDepartureTime', q12_IdealDepTime)
DesiredArrivalTime = database.DefineVariable('DesiredArrivalTime', q13_IdealArrTime)
ScheduledDelay_1 = database.DefineVariable(
    'ScheduledDelay_1',
    (DepartureTimeSensitive * (DepartureTimeMins_1 - DesiredDepartureTime))
    + (ArrivalTimeSensitive * (ArrivalTimeMins_1 - DesiredArrivalTime)),
)

ScheduledDelay_2 = database.DefineVariable(
    'ScheduledDelay_2',
    (DepartureTimeSensitive * (DepartureTimeMins_2 - DesiredDepartureTime))
    + (ArrivalTimeSensitive * (ArrivalTimeMins_2 - DesiredArrivalTime)),
)

ScheduledDelay_3 = database.DefineVariable(
    'ScheduledDelay_3',
    (DepartureTimeSensitive * (DepartureTimeMins_3 - DesiredDepartureTime))
    + (ArrivalTimeSensitive * (ArrivalTimeMins_3 - DesiredArrivalTime)),
)

Opt1_SchedDelayEarly = database.DefineVariable(
    'Opt1_SchedDelayEarly', (-(ScheduledDelay_1) * (ScheduledDelay_1 < 0)) / 60
)
Opt2_SchedDelayEarly = database.DefineVariable(
    'Opt2_SchedDelayEarly', (-(ScheduledDelay_2) * (ScheduledDelay_2 < 0)) / 60
)
Opt3_SchedDelayEarly = database.DefineVariable(
    'Opt3_SchedDelayEarly', (-(ScheduledDelay_3) * (ScheduledDelay_3 < 0)) / 60
)

Opt1_SchedDelayLate = database.DefineVariable(
    'Opt1_SchedDelayLate', (ScheduledDelay_1 * (ScheduledDelay_1 > 0)) / 60
)
Opt2_SchedDelayLate = database.DefineVariable(
    'Opt2_SchedDelayLate', (ScheduledDelay_2 * (ScheduledDelay_2 > 0)) / 60
)
Opt3_SchedDelayLate = database.DefineVariable(
    'Opt3_SchedDelayLate', (ScheduledDelay_3 * (ScheduledDelay_3 > 0)) / 60
)



import unittest
from unittest.mock import MagicMock
from biogeme.controller import Controller, CentralController
from biogeme.exceptions import BiogemeError
from biogeme.configuration import Configuration, SelectionTuple


class TestController(unittest.TestCase):
    def setUp(self):
        self.controller_name = 'my_controller'
        self.specification_names = ['spec_1', 'spec_2']
        self.the_controller = Controller(
            controller_name=self.controller_name,
            specification_names=self.specification_names,
        )

    def test_constructor(self):
        self.assertEqual(self.the_controller.controller_name, self.controller_name)
        self.assertListEqual(
            self.the_controller.specification_names, self.specification_names
        )
        self.assertEqual(self.the_controller.current_index, 0)
        self.assertListEqual(
            list(self.the_controller.dict_of_index.keys()), self.specification_names
        )
        self.assertListEqual(self.the_controller.controlled_catalogs, [])
        self.assertEqual(self.the_controller.controller_size(), 2)

    def test_all_configurations(self):
        all_configurations = self.the_controller.all_configurations()
        expected_result = ['my_controller:spec_1', 'my_controller:spec_2']
        self.assertListEqual(all_configurations, expected_result)

    def test_name(self):
        name = 'spec_2'
        self.the_controller.set_name(name)
        current_name = self.the_controller.current_name()
        self.assertEqual(name, current_name)

        incorrect_name = 'incorrect_name'
        with self.assertRaises(BiogemeError):
            self.the_controller.set_name(incorrect_name)

    def test_index(self):
        self.the_controller.set_index(1)
        current_name = self.the_controller.current_name()
        expected_name = 'spec_2'
        self.assertEqual(expected_name, current_name)
        self.the_controller.reset_selection()
        current_name = self.the_controller.current_name()
        expected_name = 'spec_1'
        self.assertEqual(expected_name, current_name)

    def test_modify_controller(self):
        self.the_controller.reset_selection()
        self.the_controller.modify_controller(step=1, circular=False)
        self.assertEqual(self.the_controller.current_index, 1)
        self.the_controller.modify_controller(step=1, circular=False)
        self.assertEqual(self.the_controller.current_index, 1)
        self.the_controller.modify_controller(step=1, circular=True)
        self.assertEqual(self.the_controller.current_index, 0)
        self.the_controller.modify_controller(step=10, circular=False)
        self.assertEqual(self.the_controller.current_index, 1)
        self.the_controller.modify_controller(step=-10, circular=False)
        self.assertEqual(self.the_controller.current_index, 0)


class TestCentralController(unittest.TestCase):
    def setUp(self):
        expression_mock = MagicMock()
        expression_mock.get_all_controllers.return_value = {
            Controller("c1", ["option1", "option2"]),
            Controller("c2", ["optionA", "optionB", "optionC"]),
            Controller("c3", ["choiceX", "choiceY"]),
        }
        self.controllers = expression_mock.get_all_controllers()
        self.central_controller = CentralController(expression_mock)

    def test_number_of_configurations(self):
        # Add test cases for the number of configurations
        self.assertEqual(self.central_controller.number_of_configurations(), 12)

    def test_get_configuration(self):
        # Add test cases for get_configuration
        # Ensure the returned configuration matches the initial state
        expected_config = Configuration(
            [
                SelectionTuple(controller='c1', selection='option1'),
                SelectionTuple(controller='c2', selection='optionA'),
                SelectionTuple(controller='c3', selection='choiceX'),
            ]
        )
        self.central_controller.set_configuration(expected_config)
        self.assertEqual(self.central_controller.get_configuration(), expected_config)

    def test_set_configuration(self):
        # Add test cases for set_configuration
        # Ensure it correctly sets the controller state
        config_to_set = Configuration(
            [
                SelectionTuple(controller='c1', selection='option2'),
                SelectionTuple(controller='c2', selection='optionC'),
                SelectionTuple(controller='c3', selection='choiceY'),
            ]
        )
        self.central_controller.set_configuration(config_to_set)
        self.assertEqual(self.central_controller.get_configuration(), config_to_set)

        # Test for incorrect configuration
        with self.assertRaises(BiogemeError):
            self.central_controller.set_configuration(
                Configuration([SelectionTuple(controller='c4', selection='wrong')])
            )

    def test_prepare_operators(self):
        # Add test cases for prepare_operators
        operators = self.central_controller.prepare_operators()

        # Make sure the expected number of operators is generated
        #    Increase: 3
        #    Decrease: 3
        #    Pairs: 3 * 2 * 4 = 24
        #    Several increase: 1
        #    Several increase: 1
        #    Total: 32
        self.assertEqual(len(operators), 32)
        # Test a specific operator function, e.g., 'Increase c1'
        operator = operators['Increase c1']
        current_config = 'c1:option1;c2:optionA;c3:choiceX'
        new_config, steps = operator(current_config=current_config, step=1)
        self.assertEqual(str(new_config), 'c1:option2;c2:optionA;c3:choiceX')
        self.assertEqual(steps, 1)

    def test_two_controllers(self):
        # Add test cases for two_controllers
        # Test for NE direction
        current_config = 'c1:option1;c2:optionA;c3:choiceX'
        new_config, steps = self.central_controller.two_controllers(
            first_controller_name='c1',
            second_controller_name='c2',
            direction='NE',
            current_config=current_config,
            step=1,
        )
        self.assertEqual(str(new_config), 'c1:option2;c2:optionB;c3:choiceX')
        self.assertEqual(steps, 1)

        # Test for invalid direction
        with self.assertRaises(BiogemeError):
            self.central_controller.two_controllers(
                current_config, 'c1', 'c2', 1, 'invalid'
            )

    def test_modify_random_controllers(self):
        # Add test cases for modify_random_controllers
        # Test increasing controllers
        current_config = 'c1:option1;c2:optionA;c3:choiceX'
        new_config, steps = self.central_controller.modify_random_controllers(
            current_config=current_config, increase=True, step=2
        )
        self.assertEqual(steps, 2)

        # Test decreasing controllers
        new_config, steps = self.central_controller.modify_random_controllers(
            current_config=current_config, increase=False, step=2
        )
        self.assertEqual(steps, 2)

    def test_set_controller(self):
        # Add test cases for set_controller
        # Test setting c1 to index 1
        self.central_controller.set_controller('c1', 1)
        expected_config = Configuration(
            [
                SelectionTuple(controller='c1', selection='option2'),
                SelectionTuple(controller='c2', selection='optionA'),
                SelectionTuple(controller='c3', selection='choiceX'),
            ]
        )
        self.assertEqual(self.central_controller.get_configuration(), expected_config)

        # Test setting an unknown controller
        with self.assertRaises(BiogemeError):
            self.central_controller.set_controller('c4', 1)

    def test_decreased_controller(self):
        # Add test cases for decreased_controller
        # Test decreasing c1 by 1
        current_config = 'c1:option1;c2:optionA;c3:choiceX'
        new_config, steps = self.central_controller.decreased_controller(
            current_config=current_config, controller_name='c1', step=1
        )
        self.assertEqual(str(new_config), 'c1:option2;c2:optionA;c3:choiceX')
        self.assertEqual(steps, 1)

        # Test decreasing an unknown controller
        with self.assertRaises(BiogemeError):
            self.central_controller.decreased_controller(current_config, 'c4', 1)


if __name__ == '__main__':
    unittest.main()



"""File b01logit_bis.py

:author: Michel Bierlaire, EPFL
:date: Sun Apr  9 17:03:31 2023

 Example of a logit model.

Same as 01logit, using bioLinearUtility, and introducing some options
 and features.  Three alternatives: Train, Car and Swissmetro SP data

Note that the parameters are defined in the file 01logitBis.toml
"""
import biogeme.logging as blog
import biogeme.biogeme as bio
from biogeme import models
import biogeme.segmentation as seg
from biogeme.expressions import Beta, bioLinearUtility

from swissmetro_data import (
    database,
    CHOICE,
    GA,
    CAR_AV_SP,
    TRAIN_AV_SP,
    TRAIN_TT_SCALED,
    TRAIN_COST_SCALED,
    SM_TT_SCALED,
    SM_COST_SCALED,
    CAR_TT_SCALED,
    CAR_CO_SCALED,
    MALE,
    SM_AV,
)

logger = blog.get_screen_logger(level=blog.INFO)
logger.info('Example b01logit_bis.py')

# Parameters to be estimated
ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_TRAIN = Beta('ASC_TRAIN', 0, None, None, 0)

# We use starting values estimated from a previous run
B_TIME = Beta('B_TIME', -1.28, None, None, 0)
B_COST = Beta('B_COST', -1.08, None, None, 0)

# Define segmentations
gender_segmentation = database.generate_segmentation(
    variable=MALE, mapping={0: 'female', 1: 'male'}
)

GA_segmentation = database.generate_segmentation(
    variable=GA, mapping={0: 'without_ga', 1: 'with_ga'}
)

segmentations_for_asc = [
    gender_segmentation,
    GA_segmentation,
]

ASC_TRAIN_segmentation = seg.Segmentation(ASC_TRAIN, segmentations_for_asc)
segmented_ASC_TRAIN = ASC_TRAIN_segmentation.segmented_beta()
ASC_CAR_segmentation = seg.Segmentation(ASC_CAR, segmentations_for_asc)
segmented_ASC_CAR = ASC_CAR_segmentation.segmented_beta()

# Definition of the utility functions
terms1 = [(B_TIME, TRAIN_TT_SCALED), (B_COST, TRAIN_COST_SCALED)]
V1 = segmented_ASC_TRAIN + bioLinearUtility(terms1)

terms2 = [(B_TIME, SM_TT_SCALED), (B_COST, SM_COST_SCALED)]
V2 = bioLinearUtility(terms2)

terms3 = [(B_TIME, CAR_TT_SCALED), (B_COST, CAR_CO_SCALED)]
V3 = segmented_ASC_CAR + bioLinearUtility(terms3)

# Associate utility functions with the numbering of alternatives
V = {1: V1, 2: V2, 3: V3}

# Associate the availability conditions with the alternatives
av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

# Definition of the model. This is the contribution of each
# observation to the log likelihood function.
logprob = models.loglogit(V, av, CHOICE)


# These notes will be included as such in the report file.
USER_NOTES = (
    'Example of a logit model with three alternatives: Train, Car and'
    ' Swissmetro. Same as 01logit and '
    'introducing some options and features. In particular, bioLinearUtility,'
    ' and automatic segmentation of parameters.'
)

# Create the Biogeme object
the_biogeme = bio.BIOGEME(
    database, logprob, userNotes=USER_NOTES, parameter_file='b01logit_bis.toml'
)

# As we have used starting values different from 0, the initial model
# is not the equal probability model. If we want to include the latter
# in the results, we need to calculate its log likelihood.
the_biogeme.calculateNullLoglikelihood(av)

the_biogeme.modelName = 'b01logit_bis'
the_biogeme.saveIterations = False

# Estimate the parameters
the_biogeme.bootstrap_samples = 100
results = the_biogeme.estimate(run_bootstrap=True)

# Get the results in a pandas table
print('Parameters')
print('----------')
pandas_results = results.getEstimatedParameters()
print(pandas_results)

# Get general statistics
print('General statistics')
print('------------------')
stats = results.getGeneralStatistics()
for description, (value, formatting) in stats.items():
    print(f'{description}: {value:{formatting}}')

# Messages from the optimization algorithm
print('Optimization algorithm')
print('----------------------')
for description, message in results.data.optimizationMessages.items():
    print(f'{description}:\t{message}')

# Generate the file in Alogit format
results.writeF12(robustStdErr=True)
print(f'Estimation results in ALogit format generated: {results.data.F12FileName}')

# Generate LaTeX code with the results

results.writeLaTeX()
print(f'Estimation results in LaTeX format generated: {results.data.latexFileName}')



""" Estimate the parameters of the logit models with the generated sample

:author: Michel Bierlaire
:date: Sat Apr 15 16:32:36 2023
"""

import biogeme.biogeme as bio
from biogeme import models

from sample import SAMPLE_SIZE as N_ALT
from choice_data import database

from model import V, log_probability

# Logit model, with correction for importance sampling of alternatives
logprob = models.loglogit_sampling(V, None, log_probability, 0)

# Create the Biogeme object
the_biogeme = bio.BIOGEME(database, logprob)
the_biogeme.modelName = 'logit'

# Calculate the null log likelihood for reporting.
the_biogeme.calculateNullLoglikelihood({i: 1 for i in range(N_ALT)})

# Estimate the parameters
results = the_biogeme.estimate(recycle=True)
print(results.shortSummary())
